﻿:Class SALT
⍝ Simple APL Library Toolkit for Dyalog

    :Field readonly Public Shared Version←2.25 ⍝ 2012/10/11 List mod
    :Include SALTUtils

    ⎕USING←0⍴⊂⍴⎕IO←1 ⋄ ⎕ML←2 ⋄ ⎕WX←3

⍝ --- Private fields

    :Field Private  Shared _TRAP←(709+⍳5) 'C' '(⎕io⊃⎕dm)⎕signal⎕en'
    :Field Private  Shared CR←⎕av[4]
    :Field Private  Shared WWIN
    :Field Private  Shared WIN
    APLV←'.'⎕wg 'aplversion'
    (WIN WWIN)←∨\(,'W')'WINE'∊APLV[3]
    :Field Public   Shared FS←(1+WIN)⌷'/\'
    :Field Private  Shared LINDEL←(~WIN)↓13 10
    :Field readOnly Shared Public SALTEXT←'.dyalog'
    :Field readOnly Shared Public SALTFOLDER←'[SALT]'

    :Field Shared SettingsTable
    split←{1↓¨(s∊⍵)⊂s←(⍵∊1↑⍺)↓⍵,⍺} ⍝ we can't trust <splitOn> to be there
    SettingsTable←0 5⍴'' ⍝ name; description; registry name; default; value
    SettingsTable⍪←'compare;the comparison program to use;CompareCMD;APL;'split';'  ⍝ e.g. [ProgramFiles]BeyondCompare
    SettingsTable⍪←'editor;the editor program to use;EditorCMD;notepad;'split';'
⍝ WorkFolders are locations where files are searched - their existence is not challenged
    SettingsTable⍪←('workdir⍟the list of storage folders to use separated by ;⍟SourceFolder⍟',SALTFOLDER,'⍟')split'⍟'
⍝ Cmd Folders are locations where Spice commands are stored - their existence is not challenged
    SettingsTable⍪←('cmddir⍟the list of Spice folders (commands) to use separated by ;⍟CommandFolder⍟',SALTFOLDER,FS,'spice⍟')split'⍟'
    SettingsTable⍪←'edprompt;editor confirmation prompt;EdPrompt;1;' split';'
    SettingsTable⍪←'newcmd;new command detection;CmdDetect;auto;' split';'          ⍝ automatic or manual
    SettingsTable⍪←'debug;debug level;DebugLevel;0;0' split';'
    SettingsTable⍪←'varfmt;variables saving format;VarFmt;APL;' split';'            ⍝ only APL and XML so far

⍝ --- Utilities

    if←/⍨                 ⍝ as in '→0 if condition'
    dotVer←{(0<⍵)/'.',⍕⍵} ⍝ >0 versions appear with a dot
    ⎕fx'r←CRef'((1+13>⍎2⍴2⊃APLV)⊃'r←2⊃⎕RSI' 'r←⍎2⊃⎕NSI')


⍝ --- Public methods

    ∇ Reboot
⍝ Bootstrap loader for SALT if the AddSALT Registry Parameter is 1 or Y or unset
      :Access public shared
      →(∨/'0nN'∊getEnvir'SALT\AddSALT')/0
      ⎕SE.SALTUtils.BootSALT
    ∇

    ∇ ref←Boot file;f;txt;method;cmd;data;folder;nf;wf;sfile;line;i;⍙FUNCTION;tmp;⍙XLOAD;arg;hasarg;name;at;TARGET
      :Access Shared Public
     
      :If isHelp arg←file
          ref←'Boot filename [arg]' ''
          ref,←'Boot from a file containing instructions' '' 'Modifiers:'
          ref,←⊂'-Function   the file contains a function to be run'
          ref,←⊂'-Xload      do not run the ⎕LX'
          ref,←⊂'If a .dyapp file each line is an instruction. If .dyalog it is a function to be executed'
          ref←⊃ref
          →0
      :EndIf
      :If hasarg←326∊⎕DR file
          (file arg)←file
      :EndIf ⍝ maybe file followed by arg to fn to run
      file←fixFsep 1⊃'-function -xload'∆parse file
      ⍙FUNCTION∨←SALTEXT≡(-⍴SALTEXT)↑file ⍝ maybe we can tell by the extension
     
      txt←i←0 ⋄ nf←⍴wf←getSetting'workdir'
      :While nf≥i←i+1
          sfile←(i⊃wf)ClassFolder file
          folder←(1⊃FS splitLast sfile),FS
          sfile←sfile{⍺,(⍵≢(-⍴,⍵)↑⍺)/⍵}(1+⍙FUNCTION)⊃'.dyapp'SALTEXT ⍝ add extension if absent
          :Trap 0
              txt←splitOnNL sfile ⍝ read file as list of strings
          :EndTrap
      :Until txt≢0  ⍝ if we found the file, exit the for loop
      (⎕EM ⎕EN)⎕SIGNAL ⎕EN if txt≡0
     
      :If ⍙FUNCTION ⍝ niladic fn?
          'tmp'⎕NS'' ⋄ at←,tmp.⎕AT name←tmp.⎕FX ⍙XLOAD{0=⍺:⍵ ⋄ ⍵/⍨~1∊¨'⍎⎕LX'∘⍷¨⍵}txt
          tmp.a←(1+hasarg)⊃⍙XLOAD arg ⍝ if the fn needs an arg we supply the one given to Boot
        ⍝ Find what type of fn we are dealing with
          :Trap 6
              tmp⍎'{}',name,(0<1 2⊃at)/' a'
          :EndTrap ⍝ trap missing result
      :Else
          TARGET←'#'
          :For line :In txt                   ⍝ read each command
              (cmd data)←line splitOn1st' '
              :Select lCase cmd
              :Case 'load'
                  :If ~FS∊data
                      data←folder,data
                  :EndIf ⍝ Default folder is where "dyalogapp" file was
                  :Trap 0
                      'Loaded: ',⍕Load'"',data,'" -target=',TARGET
                  :Else
                      'Load failed for: ',data ⋄ →0
                  :EndTrap
              :Case 'run'
                  #.⍎data
              :Case 'target'
                  TARGET←data
              :CaseList '//' '',⊂,'⍝'
              ⍝ Empty line or comment skipped
              :Else
                  ⎕←'Invalid command, ignored: ',line
              :EndSelect
          :EndFor
      :EndIf
     ⍝ End Boot
    ∇

    ∇ ref←New class;hasargs;args;⍙VERSION;cmd;⎕TRAP
      :Access Shared Public
     
      :If isHelp class
          ref←'New classname [constructor arguments]' ''
          ref,←'Create an instance of class without naming it in the workspace' ''
          ref,←⊂'Modifiers:'
          ref←⊃ref,⊂'-version=       Select specific version'
          →0
      :EndIf
     
      :If hasargs←args←~isChar class
          (class args)←class
      :EndIf
     
      ⎕TRAP←_TRAP
      class←1⊃'-version='∆parse class'1'
      cmd←'"',class,'" -noname',(0≢⍙VERSION)/' -version=',⍙VERSION
      ref←#.⎕NEW(Load cmd),hasargs/⊂args
    ∇

    ∇ {ref}←{la}Load class;file;⍙TARGET;⍙NONAME;⍙NOLINK;⍙SOURCE;⍙VERSION;name;c;w;t;v;folder;ext;target;allpaths;ver;nf;i;⎕TRAP;lastv;req;needed;line;cname;src;⎕PW;⍙DISPERSE;isns;nref;fname;path;files;found;⎕ML;isvar;ns;fix;nosrc;protect;root;xml;ts
      :Access Shared Public
⍝ This function returns either a ref or name of the Loaded object or a VTV if the source was requested or several fns were defined
⍝ or a message for anything else (failure or function) preceded by ***
      :If isHelp class
          ref←'Load [path]scriptfile' '' 'Load a class in the workspace'
          ref,←'' 'Modifiers:'
          ref,←⊂'-target=Namespace      Specifies target namespace for load'
          ref,←⊂'-disperse[=nam1,nam2]  Disperse elements in the Target namespace specified'
          ref,←⊂'-noname                Only return value, do not create name'
          ref,←⊂'-nolink                Do not "link" loaded class to source file'
          ref,←⊂'-protect               Do NOT load if the name is already defined'
          ref,←⊂'-version=              Load specific version'
          ref,←⊂'-source[=no]           Return the source ("no" means fix but use no source)'
          ref←⊃ref ⋄ →0
      :EndIf
     
      ⎕TRAP←_TRAP ⋄ ⎕ML←2
      class←fixFsep 1⊃'-target= -noname -nolink -protect -source[=]no NO -disperse[=] -version='∆parse class'1L'
      allpaths←∪ClassFolder∘class¨(getSetting'workdir')∪⊂SALTFOLDER
      lastv←0≡⍙VERSION ⍝ is the last version desired?
      ⍙VERSION←⌊num ⍙VERSION
      'la'Default 0 ⋄ (protect target)←2↑la,0 ⍝ target can be specified as a ref
      protect∨←⍙PROTECT ⍝ protection can be specified 2 ways
     
⍝ Where to define the object
      :If (1≢⍙SOURCE)∧9≠⎕NC'target'
          target←CRef ⍝ this is where we were called from
      :AndIf 0≢⍙TARGET
⍝ If -NONAME is specified we may still have to bring in other nss if :require is found and
⍝ we define these where this code was called from (where else?)
          root←∨/(⊂lCase ⍙TARGET)∊'⎕se'(,'#')
          'Invalid Target namespace'⎕SIGNAL 911 if root<9≠target.⎕NC ⍙TARGET
          target←target⍎⍙TARGET
      :EndIf
     
⍝ We support patterns
      :If ∨/'?*'∊class
          :While 0<⍴allpaths
              (path t ext)←1 splitName 1⊃allpaths
⍝ It is possible a character came back as special (? or *)
              :If found←0<⍴files←{(⍱/¨'?*'∘∊¨⍵)/⍵}Dir path,'*.',ext
                  files←FS,¨∪remVerno∘remExt¨files ⍝ remove ver no
                  path←1⊃FS splitLast 1⊃allpaths
                  t←⊂∆propagate'SOURCE NOLINK NONAME VERSION'
                  ref←protect target∘Load¨'"',¨path∘,¨files,¨'.',¨(ext,'"')∘,¨t ⍝ preserve some switches
                  allpaths←⍬
              :EndIf
              allpaths←1↓allpaths
          :EndWhile
          'Pattern not found'⎕SIGNAL 922 if~found
          →0
      :EndIf
     
⍝ Locate source
      xml←src←0
      :For i :In ⍳nf←⍴allpaths ⍝ search each folder
          (folder cname)←FS splitLast i⊃allpaths
          (cname t ext)←1 splitName cname
          :If 0<⍴(v ts)←folder ListVersionsTS cname,'.',ext
              :If lastv
                  ⍙VERSION←⌈/v
              :Else
                  lastv←t=⍙VERSION←⍙VERSION{⍺+⍵×⍺≤0}t←⌈/v  ⍝ accept neg ver
                  'version not found'⎕SIGNAL 922 if~⍙VERSION∊v
              :EndIf
              :If lastv∧(∧/c∊v)∧2=⍴c←0∪t←⍙VERSION
                ⍝ Are we dealing with the last version and do we have an unnamed file?
                  t←1⍴c[⍒⊃ts[v⍳c]] ⍝ pick the most recently changed file
              :EndIf
              src←fixTabs splitOnNL name←folder,FS,cname,(dotVer t),'.',ext
              :Leave
          :EndIf
      :EndFor
      ⎕SIGNAL 22 if src≡0 ⍝ file found? then 'name' is defined also
     
      ref←src ⋄ →0 if ⍙SOURCE≡1   ⍝ exit returning source (a VTV)
     
      isns←':⍝ '∨.=t←1↑rlb 1⊃src  ⍝ is this a namespace (as opposed to a fn)?
      isvar←'⌷'=t
     
⍝ We are now ready to FIX the object.
⍝ Some items may require other namespaces to be present,
⍝ if so, we need to bring them in BEFORE we fix the class:
     
      req←'⍝∇:require '
      :For line :In isns/(∨/¨'⍝∇:'∘⍷¨src)/src
          :If 0<⍴needed←rlb(¯1+(⍴req)+1⍳⍨req⍷lCase line)↓line
          :AndIf ~''''∊needed ⍝ no quoted names
              :Trap 22
                  1 target Load folder{'='=↑p←(¯1+⍵⍳'⍝')↑⍵:⍺,1↓p ⋄ p}needed  ⍝ same location
              :Else
                  ⎕←'*** Required file ',needed,' not found'
              :EndTrap
          :EndIf
      :EndFor
     
      :If isns
          :If ⍙DISPERSE≡0 ⍝ normal Load
              fix←⍙NONAME⍱nosrc←∨/'no' 'NO'∊⊂⍙SOURCE
              :Trap 0
                  :If protect∧fix ⍝ we need to find if the name is available
                      c←':class ' ':interface ' ':namespace '
                      :For w :In src
                          :If 4>i←1⍳⍨↑¨c⍷¨⊂lCase t←rlb w
                              cname←¯1↓(t⍳' ')↑t←(⍴i⊃c)↓t
                              :Leave
                          :EndIf
                      :EndFor
                  :AndIf 0<target.⎕NC cname        ⍝ if it already exists
                      ref←'** "',cname,'" is already defined' ⋄ →0
                  :Else
⍝ Do we want a ref or no source? Then don't fix:
                      nref←ref←fix target.⎕FIX src ⍝ otherwise use the new name
                      :If nosrc
                          cname←⍙TARGET,'.',2⊃'.'splitLast⍕ref ⍝ find full target name
⍝ Starting with 12.1 we can no longer move things across using ⎕NS ⎕OR, we need to work a bit more:
                          cname CopyNs ref         ⍝ perform copy
                      :EndIf
                  :EndIf
              :Else
                  ref←'Could not bring in <',name,'>: ',⎕IO⊃⎕DM ⋄ →0
              :EndTrap
          :Else          ⍝ disperse the elements
⍝ There are 2 things to consider:
⍝ 1. the target must be there (that has been checked already)
⍝ 2. the elements needed must be in the source
              nref←0 target.⎕FIX src
              :If ⍙DISPERSE≡1
                  needed←nref.⎕NL-2.1 9 3 4 ⋄ w←0 ⍝ everything
              :Else                                             ⍝ only the names specified
                  (⍕'invalid name:',t/needed)⎕SIGNAL 911 if∨/t←1>w←↑¨nref.⎕NC¨needed←⍙DISPERSE splitOn','
              :EndIf
⍝ OK, all seem valid, bring them in
              :If protect
                  needed←(0=w)/needed
              :Else
                  target.⎕EX¨needed
              :EndIf
              ref←'*** Unable to comply because of sourced objects'
              :Trap 16 if 9∊nref.⎕NC⊃needed
                  (⍕target)⎕NS'nref.'∘,¨needed                  ⍝ bring them in
                  ref←'* ',(⍕⍴needed),' objects dispersed in ',⍙TARGET ⍝ return OK
              :EndTrap
              :Return                                           ⍝ no need to stamp data
          :EndIf
     
      :ElseIf isvar  ⍝ the result of <Load> can only be the name of the object or its rep is -noname
          nref←target(t←{1↓¯1↓⍵[⍳⍵⍳'←']}⎕IO⊃src)src
          (ref xml)←target ⍙NONAME t protect fixVar src
          →⍙NONAME/0
      :Else ⍝ this is a fn,
⍝ Find the name of the fn/op (we could parse the text but we use ⎕FX)
          ns←⎕NS'' ⋄ fname←ns.⎕FX src  ⍝ define locally
          :If t←isChar fname
          :AndIf ⍙NONAME∨protect≤0=target.⎕NC fname
              :If ⍙NONAME
                  ref←ns.⎕OR fname ⋄ →0
              :EndIf
              target.⎕EX fname
          :AndIf isChar ref←target.⎕FX ref
              nref←target fname
          :Else
              ref←(1+t)⊃('*** could not fix <',name,'>')('** "',fname,'" is already defined')
              →0
          :EndIf
      :EndIf
      v←(⍙VERSION⌈t)×¯1*t←~lastv ⍝ mark this version as a "requested previous version"
      :If ~⍙NOLINK
          SetDelta nref name v((~⍙NONAME)/⍕ref)xml
      :EndIf
     ⍝ End Load
    ∇


    ∇ tgt←{opt}Save arg;⍙VERSION;⍙NOPROMPT;⍙BANNER;savename;cmd;folder;name;filename;named;ref;t;prev;i;r;delims;⎕TRAP;there;isFn;src;fnname;usever;maxver;ns;nosource;⍙MAKEDIR;cap;but;ext;n0;⍙CONVERT;isVar;ERR;last;allver;v;xml;⍙FORMAT
      :Access Shared Public
⍝ SAVE object [to file]
⍝ opt allows to error out instead of not saving when prompted
     
      :If isHelp arg
          r←'Save object [filename]' ''
          r,←'Save object in a specific file (default same place if already SALTed)' '' 'Modifiers:'
          r,←⊂'-convert        Convert the namespace into source form if necessary'
          r,←⊂'-banner=        Add a banner at the top of the converted namespace'
          r,←⊂'-noprompt       Do not prompt for confirmation'
          r,←⊂'-makedir        Create all necessary directories'
          r,←⊂'-version[=]     Version number to save'
          r,←⊂'-format=        Save variable in specific format (APL or XML)'
          tgt←⊃r ⋄ →0
      :EndIf
     
      tgt←0 0⍴'' ⍝ assume it won't work
      :If 0=⎕NC'opt'
          opt←0
      :EndIf  ⍝ default no options
     
      ERR←900 ⍝ this is the error starting range we use if things go wrong
     
⍝ We accept strings containing a name AND a (ref string) argument
      isVar←0 ⍝ is this a variable?
     
⍝ There are 2 ways to call this fn:
⍝ Save ref ['location']    : ref is a class or ns, location is where to save (default as tagged)
⍝ Save 'object [location]' : object may be a path. If relative CRef provides the source ns.
⍝ object may be a class 2 3 4 9 non GUI or ⎕OR object
     
      :If isFn←isChar arg    ⍝ is this a normal argument?
          t←' '⍳⍨arg←rlb arg ⋄ fnname←cmd←t↓arg ⋄ arg←(t-1)↑arg
          :If '['∊⍕ref←CRef  ⍝ unnamed (hopefully not renamed) ref?
              :If 0∊t←'.'≠arg
                  ref⍎←(¯1+i←-t⊥t)↓arg ⋄ fnname←i↑arg
              :EndIf
          :Else              ⍝ named ref
              (ref fnname)←{r←⍎⍵↓⍨¯1+i←-⊥⍨⍵≠'.' ⋄ r,⊂i↑⍵}n0←(⎕IO⊃⎕NSI){'#.'≡2↑⍵:⍵ ⋄ ⍺,'.',⍵}arg
          :EndIf
          :If (t←ref.⎕NC⊂fnname)∊9.1 9.4 9.5
              ref←ref⍎fnname ⋄ isFn←0
          :Else
              :If isVar←2∊⌊t
                  isVar←~valWithRef ref⍎fnname
              :EndIf
              'Invalid object'⎕SIGNAL ERR+2 if isVar⍱isFn←t∊3.1 3.2 4.1 4.2
          :EndIf
      :Else                  ⍝ (ref name) pair: ref is a class or a ns to save
          'Invalid argument'⎕SIGNAL 5 if 2<⍴,arg
          fnname←n0←⍕⍬⍴(ref cmd)←2↑arg,⊂''   ⍝ fully qualified nspace name
      :EndIf
     
      ⎕TRAP←_TRAP
      savename←0 fixFsep 1⊃1↑'-version[=] -makedir -format= -noprompt -convert -banner='∆parse cmd'1SL'
      :If ∨/t←'='FS=¯1↑savename
          savename←((-1↑t)↓savename),'.'afterLast⍨'.',⍕fnname
      :EndIf
      usever←1≡⍙VERSION←⌊|num ⍙VERSION ⍝ -ver alone means "use version #s"
     
⍝ Rules:
⍝ If a filename  is supplied we use it else we use the one linked to the object
⍝ If a version # is supplied we use it to possibly overwrite the file
⍝ If only -version (no =) then we start using version #s if not already the case
⍝ Without it we either overwrite or continue using version #s
     
      :If named←0≠⍴savename ⍝ file name specified
          filename←(1⊃getSetting'workdir')ClassFolder savename
      :Else                 ⍝ file name NOT specified
          :If isFn
              'Fn is not SALTed'⎕SIGNAL 911 if 0∊⍴t←fnData(ref.⎕NR fnname)fnname
          :ElseIf isVar
              'Variable is not SALTed'⎕SIGNAL 911 if 0∊⍴t←varData⊂(⍕DF ref),'.',fnname
          :Else ⍝ ref
              'Save cannot be used on SALT itself'⎕SIGNAL 911 if{0::0 ⋄ ⎕THIS≡⍵}ref
              'Ref does not point to a SALTed namespace'⎕SIGNAL 911 if 9≠ref.⎕NC'SALT_Data'
              t←ref.SALT_Data
          :EndIf
          filename←t.SourceFile
      :EndIf
      (filename t ext)←splitName filename ⍝ get rid of the verno
      (folder name)←FS splitLast filename
     
⍝ Find the version number to use
      maxver←⌈/0,allver←folder ListVersions name,'.',ext
      :If usever∨(⍙VERSION≡0)∧maxver>0
          ⍙VERSION←1+maxver
      :EndIf
     
      cap←'Save ',24{⍺>⍴⍵:⍵ ⋄ '...',(3-⍺)↑⍵}filename
      :If ⍙NOPROMPT<there←⍙VERSION∊allver
          but←'Yes' 'No',(1↑opt)/⊂'Cancel'
          →0 if 2=t←msgBox cap('Confirm overwrite of version ',(fmtVersion ⍙VERSION),'?')'Warn'but
          'abort'⎕SIGNAL ERR+1 if t=3
      :EndIf
⍝ The file may not be there but the version to use is less than the highest version
      :If ⍙NOPROMPT<there<⍙VERSION<maxver ⍝ Version is LESS than highest version number
          →0 if 1≠msgBox cap('Note: highest existing version is number ',(fmtVersion maxver),'. Proceed anyway?')'Warn'
      :EndIf
     
      (last tgt)←(folder,FS,name)∘,¨(0 1/¨⊂dotVer ⍙VERSION),¨⊂SALTEXT{0∊⍴⍵:⍺ ⋄ '.',⍵}ext
      xml←nosource←0
      :If isFn
          src←remTag ref.⎕NR fnname ⋄ ref←ref fnname
      :ElseIf isVar
          :If 0≡xml←⍙FORMAT ⍝ if not set use global setting
              :If 0∊⍴t←varData⊂n0 ⍝ do we know this var?
                  xml←'xml'≡getSetting'varfmt' ⍝ no, use global setting
              :Else ⍝ yes, use last saved format
                  xml←'xml'≡t.Format
              :EndIf
          :Else
              xml←'xml'≡t←lCase xml
              '-FORMAT=APL or XML only'⎕SIGNAL 911 if⍱/'xml' 'apl'∊⊂t
          :EndIf
          :Trap nosource←0
              src←fnname xml ∆VCR ref⍎fnname
              ref←ref fnname src
          :Else
              'Unable to create source for variable'⎕SIGNAL ERR+2
          :EndTrap
     
      :Else ⍝ NS: we grab the source; if it does not exist we create one
          :If nosource←0≡src←{16::0 ⋄ ⎕SRC ⍵}ref
⍝ For the moment we bring in the code to convert the namespace
              ns←Load'[SALT]/lib/NStoScript -noname'
⍝ Make sure the ref has a proper name
              ref.⎕DF n0
              :Trap 0 ⍝ there could be a number of reasons for this to fail
                  :If ⍙BANNER≢v←0
                      :If '⍎'=1↑v←⍙BANNER
                          v←⎕FMT CRef⍎1↓v
                      :EndIf
                  :EndIf
                  src←(0 0 v)ns.ntgennscode src←ref
⍝ Convert the namespace if required
                  :If ⍙CONVERT
                      ref←ref.##.⎕FIX src ⋄ nosource←0
                  :EndIf
              :Else
                  'Unable to convert namespace'⎕SIGNAL ERR+3
              :EndTrap
          :EndIf
      :EndIf
     
      :Trap 0
          {}makeDir⍣⍙MAKEDIR⌷pathOf t←tgt
          ns←mergeTxt src
          ns PutUTF8File t
⍝ do not update current file if not the latest version
          :If (maxver≤⍙VERSION)∧tgt≢t←last
              ns PutUTF8File t
          :EndIf
      :Else
          ('Unable to create file ',t,': ',⎕IO⊃⎕DM)⎕SIGNAL 22
      :EndTrap
      →0 if nosource
      SetDelta ref tgt ⍙VERSION({0::'' ⋄ ⍵.SALT_Data.GlobalName}ref)(3↑(3×xml)↓'aplxml')
    ∇

⍝ With Save defined we write a Snap command like this:
    ∇ r←Snap arg;Path;Sw;⍙VERSION;⍙SHOW;⍙NOPROMPT;⍙BANNER;⍙FILEPREFIX;⍙CONVERT;⍙FORMAT;tref;objs;⍙CLASS;⍙LOADFN;nc;nr;⍙PATTERN;ss;diff;class;allowed;ex;sf;files;some;makefn;name;file;ws;select;clall;valid;body;header;b;t;cut;⍙MAKEDIR;lx;⎕PP;ua;i;ext;show;txt;pat;⍙VARS;savevars;varsns;vov;varsfnam;fnames;novars;notsamefn;notsamevar;unable;to;fr;new;⍙NOSOURCE;nsn;⍙REC;rec;Trd;⍙NSTOFOLDER;all
      :Access shared public
      :If isHelp arg
          r←'Snap [path] (default current workdir)' '' 'Save all unSALTed objects in path' '' 'Modifiers:'
          r,←⊂'-class=       classes of objects to save (3 4 or 9) (*)'
          r,←⊂'-convert      convert namespaces into source form if necessary'
          r,←⊂'-banner=      Add a banner at the top of each converted namespace'
          r,←⊂'-fileprefix=  filename prefix of objects to snap'
          r,←⊂'-loadfn[=]    the path (default =arg) where to store the fn to rebuild the ws'
          r,←⊂'-nosource     do not bring in the source of namespaces with the LOAD fn'
          r,←⊂'-noprompt     skip confirmation prompt'
          r,←⊂'-makedir      create any necessary folder'
          r,←⊂'-nstofolder   create a new folder for each namespace and save each object in it'
          r,←⊂'-show[=]      do not save, show what would be saved'
          r,←⊂'-pattern=     pattern of objects to save (*)'
          r,←⊂'-vars[=]      save root variables in a separate file (=name)'
          r,←⊂'-version[=]   version number to save'
          r,←⊂'-format=      save variable in specific format (APL or XML)'
          r,←⊂'* Note: to exclude simply prefix with "~"'
          r←⊃r
          →0
      :EndIf
      Path←specialName fixFsep 1⊃1↑'-Version[=] -Format= -NoPrompt -Banner= -NsToFolder[=] -NoSource -FilePrefix= -MakeDir -Convert -Pattern= -Show[=] -Class∊ ~.12439 -LoadFn[=] -Vars[=]'∆parse arg'1SL'
      '-NSTOFOLDER cannot be specified with -BANNER, -CONVERT and -LOADFN'⎕SIGNAL 911 if(⍙REC←0≢t←⍙NSTOFOLDER)∧⍙CONVERT∨0 0≢⍙LOADFN ⍙BANNER
     
      :If ⍱/0 1∊t
          tref←⍎t
      :Else
          tref←CRef
      :EndIf
      unable←r←⍬ ⍝ return list of objects saved
     
     ⍝ If the path ends with '=' we append the name of this ws
      :If '='∊¯1↑Path
          Path←(¯1↓Path),(-⊥⍨FS≠t)↑t←⎕WSID
      :EndIf
     
     ⍝ Load function: the name of the fn can be the same has its file whose path can be specified.
      :If makefn←⍙LOADFN≢0
          'Invalid ⎕LX'⎕SIGNAL 911 if 1<1↑⍴⎕FMT ⎕LX ⍝ could be a CR/NL embedded string
          :If ⍙LOADFN≡1
              ⍙LOADFN←Path
          :EndIf
          :If isRelPath ⍙LOADFN
              ⍙LOADFN←Path{⍺,(0=⍴⍺)↓FS,⍵}⍙LOADFN
          :EndIf
          ⍙LOADFN←specialName fixFsep ⍙LOADFN
          :If '='∊1↑⍙LOADFN  ⍝ if the name starts with '=' we prepend the Path
              ⍙LOADFN←Path,1↓⍙LOADFN
          :EndIf
        ⍝ The name of the file may be supplied with an extension:
          ext←'' ⋄ name←'load_ws' ⍝ default
          :If '.'∊2⊃t←FS splitLast ⍙LOADFN
              ⍙LOADFN←1⊃t ⋄ (name ext)←'.'splitLast 2⊃t
          :EndIf
          'Invalid load fn name'⎕SIGNAL 911 if 0>⎕NC name
          ⍙LOADFN←⍙LOADFN name ext
      :EndIf
     
     ⍝ We must tackle the vars first so their ns can be detected below
      :If savevars←makefn∧⍙VARS≢varsns←varsfnam←0   ⍝ only makes sense if
          t←(tref.⎕NL-2)~(⎕SI∊⍨⊂'Spice')/'Ûargs' 'Ûcmd'
          name←(~b←valWithRef¨tref⍎¨t)/t            ⍝ no var with refs and
          varsns←tref.⎕NS name                      ⍝ a LOADFN if being produced
          unable←b/t
      :AndIf savevars←0<⍴name
          varsfnam←(1+⍙VARS≡1)⊃⍙VARS'vars_ns'
      :EndIf
     
      novars←savevars ⍝ set to 1 if we're to forbid saving vars
     
     ⍝ Find objects to snap. We allow ~ to precede the class selection meaning 'exclude'
      ⍙CLASS←⍙CLASS{⍵:⍺~'~' ⋄ ⍺}ex←'~'∊⍙CLASS
      class←⍙CLASS ∆default 2 3 4 9
     ⍝ Expand to subclasses
      allowed←∊t←2.1(3.1 3.2)(4.1 4.2)(9.1 9.4 9.5)
      class←allowed∩class,∊(2 3 4 9∊class)/t
      class←allowed~⍣ex+class ⍝ ~ means exclude
      class∪←⍙REC/9.1         ⍝ make sure namespaces are included if we recurse
      class~←novars/2.1       ⍝ exclude vars if they are treated separately
     
     ⍝ This version does not allow idioms (Save won't have it) which are class 3.3 before V12.1
      all←,tref.⎕NL-class     ⍝ , in case class is empty
      all←all~'Ûcmd' 'Ûargs' 'SALT_Data',(#=tref)/⊂'SALT_Var_Data'
     
     ⍝ We can warn that some objects cannot be saved here, e.g. refed vars
     ⍝ Remove any visible member or external objects
      all/⍨←valid←0<t←tref.⎕NC all
      valid←allowed∊⍨clall←valid/t ⍝ and variables if they are to be saved thru a ns
     
     ⍝ Some external nss are sometimes seen as regular nss and must be screened out
      :If ∨/b←valid∧9=⌊clall
          valid>←b\tref{∧/' ()'∊⍨⍕⍺⍎⍵}¨b/all ⍝ this should never happen
      :EndIf
     ⍝ Some vars may be "unsavable":
      vov←valid∨savevars∧b←clall=2.1 ⍝ Valid Or Variable
      :If ∨/t←b∧vov
          (t/valid)←(t/vov)←~valWithRef¨tref⍎¨t/all
      :EndIf
      :If b←makefn>∧/vov
      :OrIf 0<⍴unable
          ⎕←'WARNING, these objects cannot be saved:',⍕unable,b/(~vov)/all
      :EndIf
     
      nr←+/select←valid∧(tref.⎕NC all)∊class
      :If (nr>0)∧0≢⍙PATTERN
          ex←'~'∊1↑pat←⍙PATTERN~'*' ⋄ nc←⌈⌿⊃,/⍴¨objs←select/all ⍝ ignore '*' for now
          →0 if ex<nc<ss←(⍴pat)-ex ⍝ skip entire process if stem>names
          :If ex⍲nc<ss             ⍝ excluding nothing is skipped
              nr←+/select\←ex≠(ss↑¨objs)∊⊂ex↓pat
          :EndIf
      :EndIf
      →0 if savevars<0=nr
     
     ⍝ Any of these have been modified or are new?
      notsamefn←{0∊⍴nr←fnData ⍵ ⍬:1 '' ⋄ (nr.CRC≢calcCRC remTag ⍵)nr.SourceFile}
      notsamevar←{0∊⍴nr←varData⊂⍵:1 '' ⋄
          (nr.CRC≢calcCRC ⍺ ∆VCR⍎⍵)nr.SourceFile}
      (diff sf)←↓[1]⊃(⌊clall){⍺=2:⍵ notsamevar(DF tref),'.',⍵ ⋄ ⍺≠9:notsamefn tref.⎕NR ⍵
          6 16::1 '' ⋄ cc←calcCRC ⎕SRC t←tref.⍎⍵ ⋄ (sf crc)←t.SALT_Data.(SourceFile CRC) ⋄ (crc≢cc)sf
      }¨all
     
      →0 if makefn⍱some←∨/diff←diff∧select ⍝ anything to do?
     
      show←⍙SHOW≢0
     
     ⍝ Save OR return names OR save and define Load fn OR show a possible Load fn
      :If some  ⍝ find file names - replace chars in names which cause problems with the OS
     ⍝ If a relative path was specified we assume it is in the current workdir
          :If isRelPath Path
              Path←(⎕IO⊃getSetting'workdir')ClassFolder Path
          :EndIf
          Path←Path,(FS=¯1↑Path)↓FS,⍙FILEPREFIX~0
     
          new←' '∧.=¨fnames←diff/sf ⍝ reuse these filenames; note empty ones
          :If 0<⍴new/r←diff/all
              t←'0123456789abcdefghijklmnopqrstuvwxyz'
              fr←t,'àáâãåäèéêëæíîïìòóôõöøùúûüñþðßçÇABCDEFGHIJKLMNOPQRSTUVWXYZÀÁÂÃÅÄÈÉÊËÆÌÍÎÏÒÓÔÕÖØÙÚÛÜÐÝÑ∆⍙'
              to←t,'aaaaaaeeeeeiiiioooooouuuunpdsccabcdefghijklmnopqrstuvwxyzaaaaaaeeeeeiiiioooooouuuudyn^=_'
              fnames←{to[fr⍳⍵]}¨all ⍝ we need to take them ALL into account
     ⍝ There may be names which will cause conflicts under OSs like Windows
              b←((3↑¨fnames)∊'com' 'lpt')∨fnames∊'aux' 'con' 'nul' 'prn'
              :If ∨/b←b∨{(⍳⍴⍵)≠⍵⍳⍵}fnames ⍝ take into account ALL names
                  fnames←fnames,¨b/¨'-',¨⍕¨b\⍳+/b
              :EndIf
              fnames←Path∘,¨new/diff/fnames,¨⊂SALTEXT ⋄ nc←⍴SALTEXT
     ⍝ There is a chance those new names are still not unique:
              :While ∨/b←fnames∊{⍺,'.',⍵}/1 0 1/⊃splitName¨sf
                  fnames←nc⌽¨(⍕¨⍳+/b),⍨¨(-nc)⌽¨fnames
              :EndWhile
              fnames←new\fnames
          :EndIf
          :If show ⍝ guess where they would go
              sf[new/diff/⍳⍴diff]←new/fnames
     ⍝ If details requested show the filename where each object will end up
              :If 'details'≡lCase⍕⍙SHOW
                  fnames←⊃splitName¨diff/sf
                  :If ∨/0 1∊b←⍙VERSION
                      fnames[;2]←(b∨×t)/¨⍕¨1+t←⍎¨'0',¨fnames[;2]
                  :Else
                      fnames[;2]←⊂⍙VERSION
                  :EndIf
                  t←{⍺,(0=⍴⍺)↓'.',⍵}/fnames
                  r←r,¨' →'∘,¨t
              :EndIf
          :Else  ⍝  ***  save and store the effective file name  ***
              :If '⍎'∊1↑⍙BANNER
                  ⍙BANNER←'⍎',(⎕IO⊃⎕NSI),'.',1↓⍙BANNER
              :EndIf
              Sw←∆propagate'VERSION NOPROMPT CONVERT BANNER FORMAT'
              {}makeDir∘pathOf⍣⍙MAKEDIR⌷Path
              rec←files←⍬ ⋄ Trd←(DF tref),'.'
              {}{0::0 ⋄ ÷⍵<900}⎕EN ⍝ reset ⎕EN? used to detect 'CANCEL' by user below
              i←⍒b←9.1≠t←tref.⎕NC r ⋄ nsn←+/b ⍝ namespaces start after this count
              (r fnames)←(⊂i)∘⌷¨r fnames
              :For i :In ⍳⍴r ⍝ we have to loop to trap Cancel appropriately
                  :If ⍙REC∧i>nsn
                      rec,←Snap(remExt i⊃fnames),Sw,' -makedir -nstofolder=',Trd,i⊃r
                      →EndFor if 901=⎕EN
                  :Else
                      :Trap 901 902
                          files,←fnames[i]{1 Save Trd,⍵,' ',⍺,Sw}¨r[i]
                      :Else
                          files,←⊂'' ⋄ →EndFor if 901=⎕EN
                      :EndTrap
                  :EndIf
              :EndFor
     EndFor:  t←+/diff ⋄ nc←t↑~0∘∊∘⍴¨files
              r←Trd∘,¨nc/r ⍝ remove unwanted or cancelled saves
⍝ NB: Some may not have been resaved if the user said so
              sf[nc/diff/⍳⍴diff]←nc/t↑files
              r,←rec ⍝ add recursed folders
          :EndIf
      :EndIf
     
⍝                   ***  Produce code to reload the ws here  ***
      :If makefn
⍝ Produce a fn to reload the entire ws
⍝ This may be impossible if the ws includes instances (e.g. GUI objects), refs or variables
          body←0⍴ws←'⍝ This will recreate ws ',⎕WSID,' as it was on ',⍕'/:'{1↓⊃,/⍺,¨⍕¨⍵}¨↓2 3⍴⎕TS
⍝ We show the names of the objects we cannot bring back if any and limit the size of the lines to display to 90
          cut←90∘BJp ⍝ LA is max width for display
          :If 0∊vov
          :OrIf 0<⍴unable
              body←'' '⍝ These objects are not recreated:','⍝',¨⍕¨cut unable,(~vov)/all
          :EndIf
⍝ Show the objects that need saving
          :If 0∊b←∊×⍴¨files←valid/sf ⍝ some filenames are missing
              ⎕←'Warning: some filenames could not be determined'
              body,←'' '⍝ These objects are missing:','⍝',¨⍕¨cut(~b)/valid/all
          :EndIf
          (files nr nc)←b∘/¨files(valid/all)(valid/clall)
⍝ Find which version they are
          :If 0<⍴files←remVerno¨t←remExt¨files
              files←files{⍺,(0<⍴⍵)/' -ver=',⍵}¨(1+⍴¨files)↓¨t
          :EndIf
⍝ We have to order the Loads according to class (e.g. Interfaces first)
          :If ∨/b←9.4=nc
              b←b\tref{⎕ML←1 ⋄ 11::1 ⋄ ⍬⍴⍴∊⎕CLASS ⍺⍎⍵}¨b/nr
          :EndIf
          files←files[⍋b+9.1 9.5⍳nc]
          ⎕PP←17
          header←'#.(⎕io ⎕ml ⎕wx ⎕ct ⎕rl ⎕pp)←',⍕tref.(⎕IO ⎕ML ⎕WX ⎕CT ⎕RL ⎕PP)
⍝ Add variables, if so desired
          :If savevars
              t←Path,'/',varsfnam ⋄ files,←⊂t,' -disperse'
              {}Save varsns(t,' -make') ⍝ and save them of course
          :EndIf
          lx←(1⌽'''⎕LX←''',(1+⎕LX='''')/⎕LX)'⍎⎕LX'
          (fr name ext)←⍙LOADFN
          {}makeDir⍣(⍙MAKEDIR>show)⌷fr
          :If (⊂ext)∊1↓¨SALTEXT'' ⍝ create a fn
              'to'⎕NS'' ⋄ t←⊂1⌽''' -target=#',⍙NOSOURCE/' -source=no'
              {}÷name≡t←to.⎕FX name ws header,body,('{''**''≡2↑⍵,0:⎕←⍵}⎕SE.SALT.Load'''∘,¨files,¨t),''(1⌽'''⎕cs #⋄⎕wsid←''',⎕WSID),lx
              :If show
                  ⎕←to.⎕VR name
              :Else
                  {}Save'to.',name,' ',fr{⍺,(⍺∧.=' ')↓⍵}FS,name,' -noprompt'
              :EndIf
          :ElseIf ext≡'dyapp' ⍝ create a boot file
              t←remVerno∘remExt¨files
              txt←(⊂ws),body,('Load '∘,¨files),(⊂1⌽'''Run ⎕cs # ⋄ ⎕wsid←''',⎕WSID),'Run '∘,¨lx
              :If show
                  ⎕←⊃txt
              :Else
                  (mergeTxt txt)PutUTF8File fr,FS,name,'.dyapp'
              :EndIf
          :Else
              'invalid extension'⎕SIGNAL 911
          :EndIf
      :EndIf
     ⍝ End Snap
    ∇

    ∇ {r}←RemoveVersions name;⍙VERSION;⍙ALL;folder;list;⎕TRAP;fc;t;cf;⍙COLLAPSE;b;allver;v;last;⍙NOPROMPT
      :Access Shared Public
     
      :If isHelp name
          r←'RemoveVersions filename' '' 'Remove one or more versions in the SALT file system' '' 'Modifiers:'
          r,←⊂'-version=   Specific version(s) to delete'
          r,←⊂'-collapse   Keep and renumber the LAST file if needed'
          r,←⊂'-all        Remove ALL versions'
          r,←⊂'-noprompt   do not ask for confirmation'
          r←⊃r ⋄ →0
      :EndIf
      ⎕TRAP←_TRAP
      name←fixFsep 1⊃'-version= -all -collapse -noprompt'∆parse name'1L'
      v←+/0≢¨⍙ALL ⍙VERSION
      ('You must specify ',(5×2>v)↓'only one of VERSION or ALL')⎕SIGNAL 911 if 1≠v
      fc←∊ ⍝ versions must be in the set specified
     
      :If cf←∨/'<≤≥>'=t←1↑⍙VERSION←rlb ⍙VERSION
          fc←⍎t ⋄ ⍙VERSION↓⍨←1
      :EndIf
      :If v←0∊b←'-'≠t←⍙VERSION
          fc←{(⍺≥1↑⍵)∧⍺≤1↓⍵} ⋄ ⍙VERSION←b\b/t
      :EndIf
      'too many versions supplied'⎕SIGNAL 911 if(cf∨v)∧(2-cf)≠⍴⍙VERSION←num ⍙VERSION
     
      folder←2⊃(name,'*')locateIn getSetting'workdir'
      name←2⊃FS splitLast name
      allver←{⍵[⍋⍵]}folder ListVersions name
      'Object not found'⎕SIGNAL(⍴allver)↓922 ⍝ object not there
      'NO version found to delete'⎕SIGNAL 922 if 0∊⍴⍙VERSION←(b←⍙ALL∨allver fc ⍙VERSION)/allver
     
    ⍝ Find if collapsing required
      ⍙COLLAPSE∧←¯1↑b
     
⍝ If collapsing the last version won't do anything we ignore it:
      :If ⍙COLLAPSE∧(t←⌈/allver)=1+last←⌈/0,allver~⍙VERSION
          ⍙COLLAPSE←0 ⋄ ⍙VERSION~←last←t
      :EndIf
      r←Forget folder name ⍙VERSION ⍙COLLAPSE last ⍙NOPROMPT
    ∇

    ∇ r←command merge arguments;na;t;pos;b
⍝ Put arguments in command string where %n is specified
      na←⍴t←,⊂⍣(1≡≡t)+t←,arguments
      arguments←t{⍵,⍺,⍵}¨⊂WWIN⍴'"'
      :If na=+/t←'%'=command ⍝ any % specified?
⍝ We accept simple % (no n after) or %⍳na (they must ALL be there)
          pos←(¯1⌽t)/command
      :AndIf ∨/b b←(na,0)=⍴pos~na↑1↓⎕D
          (t/command)←arguments[⍋⍋b×⎕D⍳pos]
          r←∊(b⍲¯1⌽t)/command
      :Else
          r←1↓⍕command arguments
      :EndIf
    ∇

    ∇ r←Explore name;cmd;⍙MODE;⍙USE;⍙PERMANENT;folder;path;⎕TRAP;v;ext;t
      :Access Shared Public
     
      :If isHelp name
          r←'Explore name' ''
          r,←'Opens external editor on source files and explorer on folders.' '' 'Modifiers:'
          r,←⊂'-mode=         to specify Normal or Maximised window size'
          r,←⊂'-use=          to specify the EXE to use'
          r,←⊂'-permanent     to remember the EXE to use in the future'
          r←⊃r ⋄ →0
      :EndIf
     
      ⎕TRAP←_TRAP
      name←fixFsep 1⊃1↑'-Mode=normal maximised -use= -permanent'∆parse name'1SL'
      r←⍬
      ⍙MODE ∆default←'Normal'
      :If ⍙PERMANENT∧⍙USE≢0
          'editor'saveSettings ⍙USE  ⍝ replace in registry
      :EndIf
      cmd←''
      :For folder :In getSetting'workdir'
          path←folder ClassFolder name
          :If isDir path
              ⎕SIGNAL 16↓⍨WWIN ⋄ cmd←(1+WIN<WWIN)⊃'explorer.exe'(⍙USE ∆default'dolphin') ⍝ under WINE use -USE=
          :ElseIf 0<⍴v←↑ListVersions/(folder name)←FS splitLast path
              (name t ext)←splitName name ⍝ the extension may not be SALTEXT
              path←folder,FS,name,(dotVer⌈/v),SALTEXT{0∊⍴⍵:⍺ ⋄ '.',⍵}ext
              cmd←⍙USE ∆default getSetting'editor'
          :EndIf
          :If cmd∨.≠' '
              ⎕CMD ⍙MODE{⍵ ⍺}⍣WIN+cmd merge path ⋄ →0
          :EndIf
      :EndFor
      'Object not found'⎕SIGNAL 922
    ∇

    ∇ r←Compare name;⍙VERSION;list;i;files;cmd;folder;v;new;old;⍙USE;⍙PERMANENT;n;⍙ZONE;⍙TRIM;text1;text2;names;s;nf;fmt;⎕TRAP;wsver;o1;nover;tmp;nsi;isns;ext;namex;⍙SYMBOLS;max
      :Access Shared Public
     
      :If isHelp name
          r←'Compare file1 [file2]' '' 'Compares 2 classes or 2 versions of a class' '' 'Modifiers:'
          r,←⊂'-version=n       compare version n to current (ws=workspace version)'
          r,←⊂'-version=n1 n2   compare version n1 to n2 (if n<0 means max-n)'
          r,←⊂'-use=            to specify the EXE to use'
          r,←⊂'-permanent       to remember the EXE to use in the future'
          r,←'' 'If no external program is used the following modifiers are accepted:'
          r,←⊂'-zone=           the zone to show around differences (default 2)'
          r,←⊂'-trim            trim ends prior to compare'
          r,←⊂'-symbols=        use these 2 symbols for delete and insert'
          r←⊃r ⋄ →0
      :EndIf
     
      ⎕TRAP←_TRAP ⋄ r←0 0⍴''
      →0 if 0∊⍴names←'-version= -use= -permanent -zone= -trim -symbols='∆parse name'2SL'
      names←fixFsep¨names
      ⍙USE ∆default←getSetting'compare'
      wsver←'ws '≡3↑lCase⍕⍙VERSION ⍝ there may be a version AFTER 'ws'
      nover←0≡⍙VERSION             ⍝ as in  -version=ws 7
      ⍙VERSION←wsver{⍺:(2∊⍴,⍵)↓⍵ ⋄ ⍵}⍙VERSION ∆default 0
      nover∨←wsver∧⍙VERSION≡,0
     
⍝ If we specify -version=ws we can also specify an object name instead of a filename
⍝ In that case we use the sourcefile
      :If wsver
          nsi←⍎n←1⊃⎕NSI ⋄ v←n,'.',s←1⊃names
          :If 9=⎕NC v,'.SALT_Data'
              nsi←⍎{(-1+⊥⍨'.'≠⍵)↓⍵}⍕v←nsi⍎⍕names
              names←,⊂v.SALT_Data.SourceFile
          :ElseIf (⎕NC v)∊3 4
              :If ~0∊⍴nf←fnData(⎕NR v)s
                  names←,⊂nf.SourceFile
              :EndIf
          :ElseIf ~0∊⍴nf←varData⊂v
              names←,⊂nf.SourceFile
          :EndIf
      :EndIf
      'too many files/versions'⎕SIGNAL 911 if∨/(2-wsver)<(⍴names),⍴,⍙VERSION
     
      :If ⍙PERMANENT
          'compare'saveSettings ⍙USE
      :EndIf ⍝ replace in registry
     
      nf←⍴files←names
      fmt←{folder,FS,name,((0<⍵)/'.',⍕⍵),'.',ext}
     
    ⍝ Find the file(s) and its versions
      :For i :In ⍳nf
          (folder name)←FS splitLast i⊃names ⋄ (name v ext)←splitName name
          namex←name,'.',ext ⍝ remove verno, if any
          :If i=1
              o1←name
          :EndIf
          :If (~wsver)∨i=nf
              folder←(getSetting'workdir')ClassFolder¨⊂folder
              n←1⍳⍨×,⊃⍴¨v←folder ListVersions¨⊂namex ⍝ look in each folder
              (namex,' not found')⎕SIGNAL 22 if n>⍴v
              (folder v)←n⊃¨folder v ⋄ max←⌈/v
              :If 1=nf  ⍝ only 1 file specified?
                 ⍝ 1 or 2 version #s may have been specified
                 ⍝ If 1 we assume comparison is to be made between V-1 and V
                  (old new)←max{⍺+⍣(⍵≤0)+⍵}¨{¯1 0+⍣(1∊⍴⍵)+⍵},⍙VERSION
⍝ The are 2 possibilities:
⍝ 1: no version was specified and we compare the last 2 versions:
                  :If nover
                      old←⌈/¯1,v~new
                  :EndIf
⍝ 2: ⍙VERSION=x was specified and we compare x (possibly 0) & last
                  ('Version not found for ',namex)⎕SIGNAL(∧/(wsver↓old,new)∊v)↓22
                  ('Versions are the same')⎕SIGNAL 911 if old=new
                  files←fmt¨old new
              :Else
                  :If nover
                      new←max
                  :Else
                      new←max{⍺+⍣(⍵≤0)+⍵}i⊃2⍴⍙VERSION
                  :EndIf
                  ('Version not found for ',namex)⎕SIGNAL(new∊v)↓22
                  files[i]←⊂fmt new
              :EndIf
          :EndIf
      :EndFor
     
      :If (wsver∨'apl'≡lCase ⍙USE)∨' '∧.=⍙USE
          :If 9≠⎕NC'compare'
              Load'[SALT]/tools\code\compare -target=',⍕⎕CS''
          :EndIf
          compare.ZONE←⍙ZONE ∆default 2 ⋄ compare.DELINS←2↑'→',⍨' '~⍨⍙SYMBOLS ∆default'-+'
          text2←splitOnNL 2⊃files ⋄ name←⍴isns←0
          :If wsver
⍝ NOTE: the comparison is reversed for objects in the ws:
⍝ we compare the ws version TO the one on file
              :If isns←':⍝ '∨.=1↑n←rlb 1⊃text2 ⍝ is this a namespace (as opposed to a fn)?
                  text1←⎕SRC nsi⍎name←{t↑⍨¯1+⌊/':⍝'⍳⍨t←t↓⍨' '⍳⍨t←t⊃⍨1⍳⍨':'=∊1↑¨t←rlb¨⍵}text2
              :ElseIf '⌷'=1↑n ⍝ var?
                  text1←,⊂name ∆VCR nsi⍎name←1↓(¯1+n⍳'←')↑n
              :Else
⍝ We must find the name of the fn to compare to
                  'tmp'⎕NS'' ⋄ name←tmp.⎕FX text2
                  text1←remTag nsi.⎕NR name
              :EndIf
          :Else
              text1←splitOnNL o1←1⊃files
          :EndIf
          :If ⍙TRIM
              (text1 text2)←rlb∘rtb¨¨text1 text2
          :EndIf
          r←(wsver+1)⊃o1(('function' 'space'⊃⍨isns+1),' <',(rtb name),'> in the ws')
          r←'Comparing ',r,CR,'     with ',(wsver/'the one in '),2⊃files
          r←r,,CR,{0∊⍴⍵:'(they are the same)' ⋄ ⍵}text1 compare.compecv text2
      :Else
          cmd←{(⍴⍵)>p←⍵⍳']':(p↑⍵)ClassFolder p↓⍵ ⋄ ⍵}⍙USE
⍝ If the command contains % file placeholder use them
          cmd←cmd merge files
          ⎕CMD cmd'Normal'
      :EndIf
     ⍝ End Compare
    ∇

⍝ List command
⍝ Specs:
⍝ List [\root]path[\filter]
⍝ If a root is supplied we use that instead of the current workdir setting
⍝ If a filter is given then path is a folder and the filter is used for filenames
⍝ and any recursive call will include the same filter
⍝ If no filter is given then path may mean a folder with filter * or a folder followed by a filename

    ∇ r←List arg;folder;d;⍙FULL;⍙RECURSIVE;⍙RAW;dirs;i;recur;files;ver;hasver;filter;f;m;wf;⎕TRAP;es;t;⍙VERSIONS;f0;ls;⍙FOLDERS;b;isfilter;rpf;⍙TYPE;tiedalready;tie;EXT;⍙EXTENSION;remext;lc
      :Access Shared Public
     
      :If isHelp arg
          r←'list [path]' '' 'Show all .dyalog files found in a path (default current workdir)' '' 'Modifiers:'
          r,←⊂'-full[=1|2]     =1 (or no value) shows full pathnames below first folder found or below ROOT'
          r,←⊂'                =2 returns full "rooted" names.'
          r,←⊂'-recursive      Recurse through folders'
          r,←⊂'-versions       List versions'
          r,←⊂'-folders        Only list folders'
          r,←⊂'-raw            Return unformatted date and version numbers'
          r,←⊂'-type           Show the type for functions and spaces (slower)'
          r,←⊂'-extension[=]   Specify which extension to look for (all=*)'
          r←⊃r ⋄ →0
      :EndIf
     
      ls←{⍵≡'*':'a'Dir ⍺,FS,⍵ ⋄ ('ad'Dir ⍺,FS,'*'),¨'af'Dir ⍺,FS,⍵,EXT}  ⍝ list fn
      lc←lCase⍣WIN
     
      :If recur←2=≡arg
          (folder filter ⍙FULL ⍙RECURSIVE ⍙VERSIONS ⍙RAW ⍙FOLDERS ⍙TYPE EXT)←arg ⍝ Internal call
          files←folder ls filter
      :Else
⍝ Initial call goes thru here. We accept (folders\)*(name∣filter)
          ⎕TRAP←_TRAP
          f←fixFsep 1⊃1↑'-full[∊]12 -recursive -versions -folders -raw -type -extension[=]'∆parse arg'1SL'
          ⍙FULL←num ⍙FULL
          (folder filter)←FS splitLast rpf←remExt(-FS=¯1↑f)↓f ⍝ drop last \ if any
     
         ⍝ Which EXTension to use
          EXT←lc{⍵↓⍨=/2↑⍵}'.',(0 1⍳⊂t)⊃SALTEXT'*'(t←⍙EXTENSION)
     
⍝ The arg can be 1:/.../dir (all is listed), 2:/.../name (name only) 3:/.../nam* (all nam... only)
⍝ The easiest way to do it is to list the folder and look for filter as a dir
     
          isfilter←∨/'?*'∊filter ⋄ wf←getSetting'workdir'
          :If ~isRelPath rpf ⍝ rooted?
              wf←⊂folder ⋄ filter←rpf←(1+⍴folder)↓rpf
          :EndIf
⍝ If a filter is supplied we know the path is in 'folder'. If not we need to figure out
⍝ if 'filter' is a filename (in which case we're all set to proceed) or another folder
⍝ in which case we need to fetch all the files in it.
          (f0 folder files)←(rpf,'*')locateIn wf
⍝ If this was a folder we need to call Dir once more to get the files
          t←lc¨4⊃files                ⍝ caseless for Windows
          :If 1∊b←t∊⊂lc filter        ⍝ our name is there?
          :AndIf (b⍳1)⊃1⊃files        ⍝ and it's a folder?
              t←(0<⍴filter)/FS,filter
              files←'a'Dir folder,t,FS,filter←'*'
              folder,←(~isfilter)/t
          :EndIf
      :EndIf
     
      folder,←FS ⋄ es←⊂''
      d←↑files ⍝ dir, ts, size, name
      dirs←(⊂'<DIR>'),(es,d⌿⊃[1]files)[;5 1 1 3]
      dirs[;2]←folder∘,¨dirs[;2]
      dirs←↓dirs
      filter,←(0∊⍴filter)/'*'
      :If ⍙RECURSIVE∧0<t←⍴dirs ⍝ do NOT enter if t=0
          r←⍬ ⋄ ⍙FULL⌈←1
          :For i :In ⍳t
              r←r,dirs[i],List(i 2⊃dirs)filter ⍙FULL ⍙RECURSIVE ⍙VERSIONS ⍙RAW ⍙FOLDERS ⍙TYPE EXT
          :EndFor
          dirs←r
      :EndIf
      files←(es,(⍙FOLDERS⍱d)⌿⊃[1]files)[;1 5 1 4 3]
      files,←files[;2] ⍝ grab a copy of the names
      :If remext←'.*'≢EXT
          files←(EXT∘≡∘lc¨(-⍴EXT)↑¨files[;2])⌿files ⍝ keep those with the wanted extension
          files[;2]↓⍨←-⍴EXT ⍝ remove extension
      :EndIf
      :If 0<1↑⍴files
          :If remext ⍝ if the extension is kept we display all files as they are
              files[;2 3]←⊃{(∧/b)<∧/⎕D∊⍨⍵↑⍨n←-⊥⍨b←'.'≠⍵:((n-1)↓⍵)(n↑⍵) ⋄ ⍵''}¨files[;2] ⍝ split version off name
          :EndIf
          :If 1<1↑⍴files
              files←files[⍒⊃files[;5];]  ⍝ Desc by timestamp
              files←files[⍋⊃files[;2];]  ⍝ Asc by name
          :EndIf
⍝ If -versions is supplied we show the versioned files wo their unversioned ones
          b←0≠∊⍴¨files[;3] ⍝ versioned files
          :If ⍙VERSIONS    ⍝ keep the versioned files for those versioned ones
              f←b∨~t∊∪b/t←files[;2] ⍝ or those wo versions
          :Else ⍝ record the NUMBER of versions as a negative value
              files[f/⍳⍴f;3]←-+/¨b⊂⍨f←{⍵≢¨¯1↓0,⍵}lCase¨files[;2]
          :EndIf
          files←f⌿files ⍝ remove unwanted versions
⍝ If -type was specified we read the first few bytes of each file to figure out
⍝ what kind of file it is. We do this AFTER removing versions to minimize disk access.
          tiedalready←⎕NNUMS ⋄ files[;⍙TYPE/1]←'?' ⋄ files[;6 2]←folder∘,¨files[;6 2]
          :For f :In ⍳⍙TYPE×1↑⍴files
              :Trap 24 25 ⍝ file tied already?
                  tie←files[f;6]⎕NTIE¨0 ⋄ t←⎕NREAD tie,82 22 0 ⋄ ⎕NUNTIE(tie∊tiedalready)↓tie
                  t←(3×⎕AV[154 211 243]≡3↑t)↓t ⍝ UTF-8 files header?
                  t←(2×⎕AV[256 134]≡2↑t)↓t ⍝ UCS-2 files header?
                  files[f;1]←⊂'Fn'
                  :If (1↑m←2↑⎕AV[⍳5]rlb t~1⍴⎕AV)∊' ⍝:'
                      files[f;1]←⊂'Space'
                  :AndIf ':'∊1↑m ⍝ can we be more specific?
                      files[f;1]←(6⍴'If' 'Cl' 'Ns')['ICNicn'⍳m[2]]
                  :EndIf
                  :If 'â⎕'≡m
                      files[f;1]←⊂'Var'
                  :EndIf
              :EndTrap
          :EndFor
      :EndIf
      files←0 ¯1↓files ⍝ remove real filename
     
      r←(↓files),dirs
     
      :If ~recur ⍝ finalize result
          r←⊃r
          :Select ↑⍙FULL
          :Case 1
              r[;2]←(1+⍴specialName f0)↓¨r[;2]
          :Case 0
              r[;2]←{2⊃FS splitLast ⍵}¨r[;2]
          :EndSelect
          :If ⍙FOLDERS
              r←r[;,2]
          :ElseIf ~⍙RAW
              r[;5]←↓fmtDate⊃r[;5]
              r[;3]←fmtVersion¨r[;3]
              r←'Type' 'Name'('Version',⍙VERSIONS↓'s')'Size' 'Last Update'⍪r
          :EndIf
      :EndIf
     ⍝ End List
    ∇

    ∇ r←PrepareClass;s
      :Access shared
      r←ResetSettings''
⍝ Check the files needed: SALT (OK), SALTUtils (OK) and Parser + Utils:
      :If 9≠⎕SE.⎕NC s←'Parser'
          (BootPath'')BootLib s
      :EndIf
      :If 9≠⎕SE.⎕NC s←'Dyalog.Utils'
          (BootPath'')BootLib s
      :EndIf
     ⍝
    ∇

    ∇ r←ResetSettings arg;i;s;⎕TRAP
      :Access shared
      r←0 0⍴''
      ⎕TRAP←0 'E' '→0' ⍝ in case SALTUtils not included
      'unknown setting'⎕SIGNAL 911 if 0∊⍴i←i/⍳⍴i←(arg≡'')∨SettingsTable[;1]∊⊂arg←arg~' '
      SettingsTable[i;5]←regSetting¨↓SettingsTable[i;3 4],0 ⍝ cannot be empty
      '⎕se.Dyalog.Callbacks'⎕NS'' ⍝ make sure it is there
      ⎕SE.Dyalog.Callbacks.ConfirmEdit←∨/'1yY'∊5⊃SettingsTable[SettingsTable[;1]⍳⊂'edprompt';]
      ⎕SE.SALTUtils.DEBUG←'01'⍳⍬⍴5⊃SettingsTable[SettingsTable[;1]⍳⊂'debug';]
    ∇

    PrepareClass  ⍝ touch-up

    ∇ r←getSetting s
      r←5⊃SettingsTable[SettingsTable[;1]⍳⊂s;]
      :If s≡'workdir'
          r splitOn←';'
      :EndIf
     ⍝
    ∇

    ∇ r←Settings arg;pairs;⍙PERMANENT;⍙RESET;vars;names;i;val;add;rem;new;empty;⎕TRAP;k;old;valid;set
      :Access Public Shared
      :If isHelp arg
          r←'Setting [Name [Value]]' '' 'Set and returns the [new] settings' '' 'Modifiers:'
          r,←⊂'-permanent      sets permanently'
          r,←⊂'-reset          load permanently saved values'
          r,←'' 'The possible settings are:' ''
          r,←↓⍕SettingsTable[;1 2]
          r←⊃r ⋄ →0
      :EndIf
      ⎕TRAP←_TRAP
      pairs←{(1⌈⍴⍵)↑⍵}'-permanent -reset'∆parse arg'2SL'
      names←SettingsTable[;1]
      valid←0<i←names strIndex pairs[1]←lCase¨pairs[1]
      'Invalid setting'⎕SIGNAL 911 if valid⍱empty←0∊⍴↑pairs
      '-Permanent and -Reset are mutually exclusive'⎕SIGNAL 911 if ⍙PERMANENT∧⍙RESET
      pairs[valid/1]←names[valid/i] ⍝ restore to full length if it wasn't
      :If ⍙RESET
          ResetSettings 1⊃pairs
      :EndIf
     
      :If 1=⍴pairs
          :If empty
              r←SettingsTable[;1 5]    ⍝ return all values
          :Else
              r←SettingsTable[i;1 5]   ⍝ return actual value
          :EndIf
      :Else ⍝ set it
          (set val)←pairs
          old←(↑SettingsTable[i;5])splitOn';'
          :Select set
          :CaseList 'cmddir' 'workdir' ⍝ special treatment for folders; expand [ENVVARS], account for ';'s
              new↓⍨←∨/(add rem)←',~'=↑new←val
              new←'\[[^][;]+\]'⎕R{specialName ⍵.Match}new
              'invalid path specified'⎕SIGNAL 922 if∨/'[]'∊new
              val←{⍵/⍨' '∨.≠¨⍵}new splitOn';' ⍝ split names that need to be and remove empties
              :If add
                  val←∪val,old         ⍝ side-effect to move ahead
              :ElseIf rem
                  val←old~val
              :EndIf
              val←val,(0∊⍴val)/⊂SALTFOLDER,(set≡'cmddir')/FS,'Spice'
            ⍝ Perform minimal validation on paths
              val←↑{⍺,';',⍵}/val ⍝ merge paths
          :Case 'newcmd'
              'new commands must be detected AUTO[matically] or MANUAL[ly]'⎕SIGNAL 911↓⍨∨/'auto' 'manual'≡¨⊂val←lCase val
          :Case 'edprompt'
              val←⍕⎕SE.SALTUtils.ConfirmEdit←∨/'1yY'∊val
          :Case 'debug'
              val←⍕⎕SE.SALTUtils.DEBUG←{∨/'yY'∊⍵:1 ⋄ +/2⊃⎕VFI ⍵}val
          :Case 'varfmt'
              'Variables format can only be APL or XML'⎕SIGNAL 911↓⍨∨/'apl' 'xml'≡¨⊂lCase val
          :EndSelect
          SettingsTable[i;5]←⊂val
          r←SettingsTable[i;1 5]       ⍝ return effective value
      :EndIf
     
      :If ⍙PERMANENT
          saveSettings/r
      :EndIf
      :If ~empty
          r←2⊃r
      :EndIf
     ⍝ End Settings
    ∇

⍝ $Revision: 14933 $ $Date: 2012-10-11 16:43:10 +0100 (Thu, 11 Oct 2012) $
:EndClass