﻿:Class EditFileAssociations

⍝∇:require =\WinReg.dyalog  

    ⎕io←1

    ∇ r←List
      :Access Shared Public
      r←⎕NS''
      r.Name←'EFA'
      r.Desc←'Edit APL file associations'
      r.Group←'System'
      r.Parse←''
    ∇

    ∇ {rc}←Run(Cmd Args);⎕USING
      :Access Shared Public
      :Select Cmd
      :Case 'EFA'
          RunEdit
          rc←⍪''
      :EndSelect
    ∇

    ∇ r←Help Cmd
      :Access Shared Public
      :Select Cmd
      :Case 'EFA'
          r←'Changes the associations of Dyalog Application (.dyapp files)',CR←⎕AV[4]
          r,←'and Dyalog Workspace files (.dws files)'
      :EndSelect
    ∇

    :Section ControlFunctions

    ∇ {applied}←RunEdit;path;curr;ivers;ipaths;f;text;⎕IO;⎕ML
      applied ⎕IO ⎕ML←0 1 1
      :If IsNotWindows
          ⎕←'This utility applies only to Windows versions of Dyalog APL'
          :Return
      :EndIf
     
      text←⊂'This utility allows you to specify which installed version of Dyalog APL'
      text,←⊂'is to be associated with Dyalog files, specifically workspaces (.dws) and'
      text,←⊂'Dyalog applications (.dyapp).'
     
      'File Associations Management Utility'MsgBoxAlert text
      :If ~IsUserAnAdmin
          'Access Error'MsgBoxAlert'This utility requires Dyalog to be run with administrative privileges'
          :Return
      :EndIf
     
      ivers←InstalledVersions ⍝ A vector of all installed versions of Dyalog APL from 11.0 onward
      ipaths←##.WinReg.GetString¨'HKEY_CURRENT_USER\'∘,¨ivers,¨⊂'\dyalog' ⍝ A vector of the installation directories of installed versions
      curr←CurrentVersionUsed
      MakeGui ⍝ Generates form
   
      ⎕DQ'f'
    ∇

    ∇ (rc msg)←CurrentVersionUsed;r;path;exists;ind;mask
    ⍝ Determines which version of Dyalog APL, from the list of versions installed on the machine, is the
    ⍝ version already associted with files with the extesions ".dws" and ".dyapp".
      rc msg←(0 0)('' '')
     
      rc←##.WinReg.DoesKeyExist¨'HKEY_CLASSES_ROOT\'∘,¨'dwsfile' 'dyappfile'
      :If 0∊rc
          ((~rc)/msg)←⊂'Extension not defined in registry'
      :EndIf
     
      r←##.WinReg.GetString¨'HKEY_CLASSES_ROOT\'∘,¨('dwsfile' 'dyappfile'),¨⊂'\shell\open\command\'
     
      exists←Exists¨{({⍵/⍳⍴⍵}¯3⌽'.exe'⍷⍵)↑⍵}¨r
      ((exists<rc)/msg)←⊂'The version of Dyalog APL associated with this file may have been moved or uninstalled.'
      path←{(⌽≠\<\'\'=⌽⍵)/⍵}¨r
      ind←ipaths⍳path
      mask←((⍴ipaths)≥ind)∧(exists∧rc)
      msg←msg,¨mask\ivers[mask/ind]
    ∇

    ∇ Apply;ind;selpath;selver;vernum;mask 
    ⍝ Identifies the versions of Dyalog installed on the machine
    ⍝ Prompts the user to select one of these versions to associate
    ⍝ .dws and .dyapp files.
      mask←'f.list'⎕WG'SelItems'
      :If ~∨/mask
          'Edit File Associations' MsgBox 'Please select a version of Dyalog APL.'
      :Return
      :Endif
      ind←{⍵/⍳⍴⍵}mask
      selpath←ind⊃ipaths
      selver←ind⊃ivers
      :If ~'EditFileAssociations'MsgBoxYN'Associate extensions with ',(AddClassic 16↓selver),'?'
          :Return
      :EndIf
      vernum←GetVersion selver
      :If applied←Set selpath selver
          'Edit File Associations'MsgBoxAlert'Association successful'
      :EndIf
      ⎕EX'f'
    ∇

    ∇ rc←Set(path vernum);types;str;bin;mask;Text;dws;dyapp
    ⍝ Builds data to be written to registry
    ⍝ Writes to registry.
      dws dyapp←⊃curr
      str←0 2⍴''
     
      str⍪←'.dyalog\' 'dyalogfile'
     
      :If dws
          str⍪←'.dws\' 'dwsfile'
          str⍪←'dwsfile\' 'Dyalog APL Workspace'
          str⍪←'dwsfile\shell\Open\command\'(path,'dyalog.exe "%1"')
          str⍪←'dwsfile\shell\open (do not start)\command\'(path,'dyalog.exe -x "%1"')
          str⍪←'dwsfile\Version\'vernum
      :EndIf
     
      :If dyapp
          str⍪←'.dyapp\' 'dyappfile'
          str⍪←'dyappfile\' 'Dyalog APL Script Application'
          str⍪←'dyappfile\shell\Open\command\'(path,'dyalog.exe DYAPP="%1"')
          str⍪←'dyappfile\Version\'vernum
      :EndIf
     
      bin←0 2⍴'' 0
     
      :If dws
          bin⍪←'dwsfile\shell\open\EditFlags' 1
      :EndIf
     
      bin⍪←'dyalogfile\shell\open\EditFlags' 1
     
      :If dyapp
          bin⍪←'dyappfile\shell\open\EditFlags' 1
      :EndIf
      ⍝ Format the values
      str[;1]←(⊂'HKEY_CLASSES_ROOT\'),¨str[;1]
      bin[;1]←(⊂'HKEY_CLASSES_ROOT\'),¨bin[;1]
     
      ##.WinReg.PutString¨↓str ⍝ Sets all String Values
      ##.WinReg.PutBinary¨↓bin ⍝ Sets all Binary Values
     
      ⍝ Check for Completion
     
      mask←str[;2]≢¨##.WinReg.GetString¨str[;1]
      mask,←∊bin[;2]≠¨##.WinReg.GetValue¨bin[;1]
     
      :If 1∊mask
          Text←'The following registry keys',⊂⎕UCS 13 10
          Text,←'were not properly written:',⊂⎕UCS 13 10
          Text,←(mask)/(str⍪bin)[;1],¨⊂⎕UCS 13 10
          'Errors'MsgBoxAlert∊Text
          rc←0
      :Else
          rc←1
      :EndIf
    ∇
           
    :EndSection

    :Section GUI
     
    ∇ MakeGui;Text;Y;X;vers;size
    ⍝ Creates GUI
      vers←16↓¨ivers
     
      Text←⊂'The list below contains the versions of Dyalog APL'
      Text,←⊂'installed on this machine.'
      Text,←⊂''
      :If (=/1⊃curr)∧(≡/2⊃curr)
          Text,←⊂'Dyalog Workspace (.dws) Files and Dyalog Application'
          Text,←⊂'(.dyapp) Files are both associated with: '
         Text,←⊂AddClassic 16↓1⊃2⊃curr
      :Else
      :If 1⊃1⊃curr
          Text,←⊂'Dyalog Workspace (.dws) Files are currently associated'
          Text,←⊂'with: ',AddClassic 16↓1⊃2⊃curr
      :Else
          Text,←⊂'Dyalog Workspace (.dws) files are not currently associated'
          :If 'Edit File Associations'MsgBoxYN'Dyalog Workspace (.dws) files are not currently associated' 'with any executables. Would you like to create those registry keys?'
              Text,←⊂'with any executable, but will be on set.'
          :Else
              Text,←⊂'with any executable, but will not be set.'
          :EndIf
     
      :EndIf
      Text,←⊂''

      :If 2⊃1⊃curr
     
          Text,←⊂'Dyalog Application (.dyapp) Files are currently'
          Text,←⊂'associated with: ',AddClassic 16↓2⊃2⊃curr
      :Else
          Text,←⊂'Dyalog Application (.dyapp) files are not currently associated'
          :If 'Edit File Associations'MsgBoxYN'Dyalog Application (.dyapp) files are not currently associated' 'with any executables. Would you like to create those registry keys?'
              Text,←⊂'with any executable, but will be set to the selected item.'
          :Else
              Text,←⊂'with any executable, but will not be set.'
          :EndIf
      :EndIf
      :EndIf
      Text,←⊂''
      Text,←⊂'Select the version you would like to associate with both'
      Text,←⊂'.dws and .dyapp files and click apply.'
      listx←300⌊19×⍴vers
      size←(250+listx),365
      'f'⎕WC'Form' 'Edit Dyalog File Associations'('Size' size)('Coord' 'Pixel')('Sizeable' 0)('MaxButton' 0)('MinButton' 0)
      'f.fnt'⎕WC'Font'('Pname' 'Arial')('Size' 16)
      'f'⎕WS'Event'('Close' 'Cancel')
      'f'⎕WS'FontObj' 'f.fnt'
      'f.msg'⎕WC'Text'(↑Text)(15 15)
      'f.list'⎕WC'List' ('Items' (AddClassic¨vers)) ('Posn' (200 15)) ('Size' (listx,335))
     
      'f.bapply'⎕WC'Button'('Caption' 'Apply')('Posn' ((1⊃size)-40) 100)('Size' 30 70)('Event' 'Select' 'Apply')
      'f.bcancel'⎕WC'Button'('Caption' 'Quit')('Posn' ((1⊃size)-40) 200)('Size' 30 70)('Event' 'Select' 'Cancel')
      ⎕NQ'f' 'Gotfocus'
    ∇
      
    ∇ {r}←{title}MsgBox msg;m
    ⍝ Creates a message box
      :If 0=⎕NC'title' ⋄ title←'' ⋄ :EndIf
      'm'⎕WC'MsgBox'title msg
      r←⎕DQ'm'
    ∇

    ∇ {r}←Cancel;M;Z
      :If 'Edit File Associations'MsgBoxYN'Are you sure you want to quit?'
          ⎕EX'f' 'ivers' 'ipaths'
      :EndIf
    ∇

    ∇ title MsgBoxAlert msg;mb
      :If 0=⎕NC'title' ⋄ title←'' ⋄ :EndIf
      'mb'⎕WC'MsgBox'('Caption')('Text'msg)('Style' 'Warn')('Btns' 'OK')
      {}⎕DQ'mb'
    ∇

    ∇ {r}←{title}MsgBoxYN msg;mb
      :If 0=⎕NC'title' ⋄ title←'' ⋄ :EndIf
      'mb'⎕WC'MsgBox'('Caption'title)('Text'msg)('Style' 'Query')('Event'('MsgBtn1' 'MsgBtn2')1)
      r←'MsgBtn1'≡2⊃(⎕DQ'mb'),2⍴⊂''
    ∇
    :EndSection

    :Section Utilities

    ∇ r←Exists name
      ⍝ Does file exist?
      r←1
      :Trap 19 22
          ⎕NUNTIE name ⎕NTIE 0
      :Else
          r←0
      :EndTrap
     
    ∇
    AddClassic←  {'Unicode'≢¯7↑⍵:⍵,' Classic' ⋄ ⍵}

    ∇ r←IsUserAnAdmin;IsUserAnAdmin
     ⍝ check if this process is being "Run as Administrator"
     ⍝ BPB - nov2010
      :Trap r←0
          r←⍎⎕NA'I Shell32|IsUserAnAdmin'
      :EndTrap
    ∇
  
    ∇ r←InstalledVersions;sk;mask
      sk←##.WinReg.GetAllSubKeyNames'HKEY_CURRENT_USER\Software\Dyalog'
      mask←1∊¨'Dyalog APL/'∘⍷¨sk
      r←(⊂'Software\Dyalog\'),¨mask/sk
    ∇

    ∇ r←IsNotWindows
      r←'Window'≢6↑⊃'.'⎕WG'APLVersion'
    ∇

    ∇ r←IsUnicode
      r←82≠⎕DR''
    ∇

    GetVersion←{⍕10×⊃(//)⎕VFI ⍵}

    :EndSection

:EndClass