﻿:Namespace wsc

    ⍝ *** Objects comparison utility made from ws 'compare11'

    (⎕IO ⎕ML)←1

   ⍝ All those switches except GATHEROUTPUT and EXSTRING apply to wscompare as well as varcompare because it compares
   ⍝ whole namespaces which may contain fns. cfcompare has the same switches as varcompare plus -cpts
   ⍝ fncompare only has a few switches
    sw←,⊂'F-DELINS=:      the 2 characters to denote deletion and insertion'
    sw,←⊂'B-EXNAMES=:     exclude those items from the comparison'
    sw,←⊂'1-EXSTRING=:    exclude items with those strings in them'
    sw,←⊂'F-EXTS:         exclude the function timestamps'
    sw,←⊂'1-GATHEROUTPUT: gather and return all output'
    sw,←⊂'B-NLINES=:      the max number of lines of display of a variable'
    sw,←⊂'F-NOLASTLINE:   remove the tag line, if any, before comparison'
    sw,←⊂'F-NORMALIZE:    remove excesss blanks before comparing'
    sw,←⊂'B-NSSRC:        force use of source for namespaces if they exist'
    sw,←⊂'B-SHOW=:        which sections to show'
    sw,←⊂'F-ZONE=:        the number of lines to show before and after each difference'
    sw,←⊂'8-CPTS=:        list of components to compare'

    ∇ s←lCase s;b;⎕IO;i;n;l;u
      ⍝ Lowercase string
      (l u)←'abcdefghijklmnopqrstuvwxyzàáâãåèéêëòóôõöøùúûäæü' 'ABCDEFGHIJKLMNOPQRSTUVWXYZÀÁÂÃÅÈÉÊËÒÓÔÕÖØÙÚÛÄÆÜ'
      n←⍴l ⋄ →(∨/b←n>i←u⍳s)↓⎕IO←0
      (b/s)←l[b/i]
    ∇

    cSw←lCase¨allSw← sw←(¯1+sw⍳¨':')↑¨SwitchesDefn←sw
    Sets←↓⊖2 2 2 2⊤¯1+(⎕D,⎕A)⍳⍬∘⍴¨SwitchesDefn

    cut←{⍺←1 ⋄ ⍺↓¨(⍵∊1↑⍵)⊂⍵}
    Cmds←'wscompare'  'varcompare'  'fncompare'   'cfcompare'

    ∇ r←List
      r←⎕NS¨(⍴Cmds)⍴⊂⍬
     ⍝ Name, group, short description and parsing rules
      r.Name←Cmds
      r.Group←⊂'wsutils'
      r.Desc←,∘' comparison'¨'workspaces' 'variables' 'functions' 'component files'
      r.Parse←'2S' '2' '2' '2',¨⍕¨Sets/¨⊂1↓¨cSw
    ∇

    ∇ r←Run(Cmd Args)
      ZONE←_ZONE Args.Switch'zone'
      SHOW←_SHOW Args.Switch'show'
      NLINES←_NLINES Args.Switch'nlines'
      DELINS←_DELINS Args.Switch'delins'
      (NOLASTLINE NSSRC EXSTRING EXNAMES EXTS TRIM)←Args.Switch¨'nolastline' 'nssrc' 'exstring' 'exnames' 'exts' 'normalize'
     
      :Select Cmd
      :Case 'wscompare'
          r←compare Args
      :Case 'fncompare'
          'Invalid function name'⎕SIGNAL 11 if~∧/(##.THIS.⎕NC↑Args.Arguments)∊3 4
          r←##.THIS comp2fn Args
      :Case 'varcompare'
          'Invalid variable name'⎕SIGNAL 11 if~∧/(##.THIS.⎕NC↑Args.Arguments)∊2 9
          r←##.THIS comp2var Args
      :Case 'cfcompare'
          r←compfiles Args
      :EndSelect
    ∇

    ∇ r←lev Help Cmd;⎕ML;sw;i;obj
      :Access Shared Public
      sw←⍴⎕ML←1 ⋄ i←Cmds⍳⊂Cmd
      r←1↓∊CR,¨(i⊃Sets,1)/2↓¨SwitchesDefn
      obj←i⊃'workspace' 'variable' 'function' 'file' 'object'
      r←'Compare 2 ',obj,'s in details. Switches are:',CR,r
      :If lev>0
          r←Describe
      :EndIf
    ∇

    CR←⎕av[⎕io+3]

    _←{⍵,CR,(∨\⍵≠' ')\'='}'                         Compare utilities'
    _,←CR
    _,←CR,'      The following commands will show the difference between 2 objects in the current'
    _,←CR,'namespace:'
    _,←CR
    _,←CR,' ]wscompare w1 w2      compares 2 workspaces'
    _,←CR,' ]fncompare f1 f2      compares 2 fns; f1 and f2 are the names of fns'
    _,←CR,' ]varcompare v1 v2     compares 2 variables; v1 (or v2) may be a namespace'
    _,←CR,'                       Enclosed arrays are detailed'
    _,←CR,'                       only if they are vectors of same shape.'
    _,←CR,'                       Differences are shown on simple objects of'
    _,←CR,'                       rank 2 or less or if they are small enough.'
    _,←CR,' ]cfcompare f1 f2      compares 2 component files'
    _,←CR,'                       same rules as varcompare for each component'
    _,←CR
    _,←CR,'    The format of FNcompare is as follows: the function on the left is'
    _,←CR,'assumed to be the one which has changed. the fn on the right is assumed'
    _,←CR,'to be the new function. The output is based on these assumtions. Each'
    _,←CR,'line of the fn on the left is preceded by its line number [xx]. Where'
    _,←CR,'the line has been deleted a symbol (like ←) appears before the line'
    _,←CR,'number. Where a new line has been inserted there is no line numbering'
    _,←CR,'and a symbol (like →) will appear.'
    _,←CR
    _,←CR,'example:  ∇fna                  ∇fnb'
    _,←CR,'       [1] same line         [1] same line'
    _,←CR,'       [2] fna line 2        [2] fnb line 2'
    _,←CR,'       [3] same line 3       [3] same line 3'
    _,←CR,'       [4] ⍝ comment deleted [4] new common line'
    _,←CR,'       [5] new common line   [5] ⍝ new comment'
    _,←CR,'          ∇                     ∇'
    _,←CR
    _,←CR,'      ]fncompare  fna fnb'
    _,←CR,'-[0]  fna'
    _,←CR,'+     fnb'
    _,←CR,' [1]  same line'
    _,←CR,'-[2]  fna line 2'
    _,←CR,'+     fnb line 2'
    _,←CR,' [3]  same line 3'
    _,←CR,'-[4]  ⍝ comment deleted'
    _,←CR,' [5]  new common line'
    _,←CR,'+     ⍝ new comment'
    _,←CR
    _,←CR,'Note that in order to avoid lengthy output of big functions with little'
    _,←CR,'difference only sections of the functions are shown, those where the changes'
    _,←CR,'occurred plus a couple of lines before and after (see switch ZONE).'
    _,←CR
    _,←CR,'Note: this version does not report the stack state'
    _,←CR,'      and may not work fine for very large objects'
    _,←CR,'      ⎕pw is used to delimit output'
    _,←CR
    _,←CR
    _,←CR,'Switches'
    _,←CR,'      You can use the following switches:'
    _,←0 Help ⍬
    _,←CR
    _,←CR,'--- Output specifications; switch SHOW'
    _,←CR
    _,←CR,'      The amount of info displayed is determined by a few settings. One of'
    _,←CR,'them, -SHOW, contains 3 values for each one of the SECTIONS of the report,'
    _,←CR,'that is the 2 "unique symbols" sections and the "common symbols" section,'
    _,←CR,'i.e.'
    _,←CR,'    - the symbols found ONLY in the first object/ws'
    _,←CR,'    - those found ONLY in the second one'
    _,←CR,'    - those found in both'
    _,←CR
    _,←CR,'      each section has 3 PROPERTIES:'
    _,←CR,'    - the number of symbols'
    _,←CR,'    - their names'
    _,←CR,'    - their contents (or difference)'
    _,←CR
    _,←CR,'      The first property is always displayed. The last 2 depend on the'
    _,←CR,'section value in SHOW:'
    _,←CR,'   if 0 only the number of objects is shown'
    _,←CR,'      1 the names are also displayed'
    _,←CR,'      2 their contents/difference is also displayed'
    _,←CR,'      3 both are also displayed'
    _,←CR
    _,←CR,'      The default value is 1 1 3, that is the first 2 sections have'
    _,←CR,'their names displayed only and the third one shows names and differences.'
    _,←CR,'To change temporarily this setting use -show=N, for example:'
    _,←CR
    _,←CR,'      ]wscompare ws1 ws2 -show=2 ⍝ to show names only for all sections'
    _,←CR
    _,←CR,'      Another switch affecting display is NLINES which shows the first'
    _,←CR,'N lines of a variable in the "unique symbols" area.'
    _,←CR
    _,←CR,'      Yet another one is ZONE which affects the display of differences'
    _,←CR,'between objects. This is the number of lines shown before and after'
    _,←CR,'any difference found.'
    _,←CR
    _,←CR,'--- Files ---'
    _,←CR
    _,←CR,'      To compare 2 files use'
    _,←CR,'      ]cfcompare  ''file1''  ''file2'''
    _,←CR,'This will compare each component with the same number. That is, if file1'
    _,←CR,'ranges from components 10 to 20 and file2 ranges from 15 to 45 then only'
    _,←CR,'components 15 to 20 will be compared.'
    _,←CR,'If a files uses a lock specify it after the name and a colon, e.g.'
    _,←CR,'      ]cfcompare  fileabc:123   filedef:4321'

    Describe←_

    DELINS←_DELINS←'-+'   ⍝ delete/insert characters

    MAXDIMS←25 160 ⍝ maximum size of the display of a veriable

    NLINES←_NLINES←10

    PTYPE←'namespace'

    SHOW←_SHOW←1 1 3

    vtom ←{↑(⍵=1↑⍵)⊂⍵}

    SYSVARS←vtom '⎕lx ⎕ct ⎕io ⎕pp ⎕ml ⎕rl ⎕div ⎕trap ⎕using'

    Version←2.50        ⍝ based on V2.47

    ZONE←_ZONE←2        ⍝ zone to display before and after matches

    NSSRC←TRIM←NOLASTLINE←EXSTRING←0

    ⎕ex '_'

    addDiff←{'≠'≠1↑⍵:'' ⋄ '≠',⍺,CR,1↓⍵}

    ∇ r←{a}cat1 b;sa;sb
     ⍝ danb 83/10/12 catenates 2 objets of ≠ ⍴s on 1st axis
     ⍝ if args are empty vectors they're replaced by 0 0⍴''
     ⍝ ⍴ won't use up more space if the same but ↑ will
      r←0,1↓(⍴a←(sa←¯2↑1,(×⍴,a),⍴a)⍴a)⌈⍴b←(sb←¯2↑1,(×⍴,b),⍴b)⍴b
      →(1↓sa=sb)⍴l10 ⍝ skip ↑ ?
      a←(r⌈sa)↑a ⋄ b←(r⌈sb)↑b
     l10:r←a⍪b
    ∇

    ∇ r←{a}cat2 b;sa;sb
     ⍝ danb 2003 catenates 2 objets of ≠ ⍴s on 2nd axis
     ⍝ if args are empty vectors they're replaced by 0 0⍴''
     ⍝ ⍴ won't use up more space if the same but ↑ will
      r←1 0×(⍴a←(sa←¯2↑1,(×⍴,a),⍴a)⍴a)⌈⍴b←(sb←¯2↑1,(×⍴,b),⍴b)⍴b
      →(1↑sa=sb)⍴l10
      a←(r⌈sa)↑a ⋄ b←(r⌈sb)↑b
     l10:r←a,b
    ∇

    ∇ d←p1 compack p2;n1;n2;c1;c2;i;obj;m;n;⎕IO;cr2;show;h;l;c;O;fullpath;pathcopy;any;NL;Z;t;diff;head;srcs;T;s1;s2
     ⍝ danb 840531 compare two packages/refs, adapted for classes 2008
     ⍝ p1 & p2 are either refs, ⎕ORs or char vectors (⎕SRCs); returns  '[≠]text of changes'
     ⍝ Objects need to be similar to be compared (a ⎕OR of a NS with another ⎕OR of a NS)
     ⍝
      ⎕IO←1 ⋄ show←,2 2⊤3⍴SHOW
      Z←0⍴⎕FX'O s' '→(⍴s)↓0⋄Z←Z,,CR,s'           ⍝ use a fn to gather output
     
     ⍝ Check if there are obvious differences
      l←1+isor¨p1 p2 ⋄ d←'≠',⊃{⍺,' vs ',⍵}/l⊃¨⊂'NS' '⎕OR'
      :If 2∧.=l ⍝ both ⎕ORs
          d←(0<⍴d)↓'≠⎕ORs: ',d←(⎕NS p1)compack ⎕NS p2
      :Else
          →0 if≠/l
          d←'≠',⍕{⍺,' vs ',⍵}/T←(n1 n2)←nsType¨p1 p2 ⍝ exit if different types (≠ for forms etc.)
          →0 if≢/T~¨':' ⍝ remove Source indication
         ⍝ Fix namespaces locally. Other spaces (eg Class) are NOT fixed
         ⍝ because they may require other elements (based, include)
          :If ∨/d←':namespace'∘≡¨lCase¨T lev t←p1 p2
              (d/t)←(⎕NS''){⍺.⎕FIX ⍵}¨d/t
              p1 p2←t
          :EndIf
          d←'≠Instances?'
          →0 if~∨/t←(⎕NC⊂'p1')=2.1 9.1 9.4 9.5       ⍝ or if of an impossible type to show
     
     ⍝ Here we have 2 objects of the same type.
          :If srcs←∨/2↓t ⍝ compare sources of class/interfaces
          :OrIf srcs←NSSRC∧∧/{11 16::0 ⋄ 1{⍺}⎕SRC ⍵}¨p1 p2
          :OrIf ∧/isVTV¨p1 p2
              :If ~srcs
                  n1←1↓{(⍵⍳' ')↑⍵},⎕IO⊃p1
              :EndIf
              d←(0∊⍴d)↓'≠Objects type: ',n1,(¯1+d⍳CR)↓d←1↓(⎕SRC⍣srcs+p1)compvar ⎕SRC⍣srcs+p2
     
          :Else ⍝ could be ⎕ORs or Namespaces
              Z←1⌽'sObjects are ',1⊃T
              :If 1↑t ⍝ ⎕ORs?
                  :Trap 11
                      (p1 p2)←⎕NS¨p1 p2
                  :Else
                      d←'≠Unknown type: ',1⊃⎕DM ⋄ →0
                  :EndTrap
              :EndIf
              n1←NLnoSALT p1 ⋄ n2←NLnoSALT p2
     
     ⍝ Since we need to keep track of nested package we have to use a global.
              :If 0=⎕NC'Path' ⋄ ⎕SHADOW'Path' ⋄ Path←'' ⋄ :EndIf
              pathcopy←Path ⋄ fullpath←Path,(0∊⍴Path)↓'.'
     
     ⍝ If we have 2 sourced nss we could compare the scripts instead.
     ⍝ List objects in left but not in right
              obj←' outstanding objects in '
              :If any←∨/i←0=c←×n2 rowfind n1
                  O'*** ',(⍕+/i),obj,'1st'
                  :If show[4]
                      O v∆m fit i⌿n1
                  :EndIf
                  :If show[1]
                      O∘(p1∘displayfp)¨↓i⌿n1
                  :EndIf
              :EndIf
     
     ⍝ List objects in right but not in left
              n1←c⌿n1 ⍝ keep common objects only
              :If ∨/i←0=n1 rowfind n2
                  m←(3××⍴Z)⍴CR ⋄ any←1
                  O m,'*** ',(⍕+/i),obj,'2nd'
                  :If show[5]
                      O v∆m fit i⌿n2
                  :EndIf
              :AndIf show[2]
                  O∘(p2∘displayfp)¨↓i⌿n2
              :EndIf
     ⍝ Display common objects. Wait for the list until we know what's different.
              m←1⍴⍴obj←n1 ⋄ diff←l←⍳i←0 ⋄ cr2←2⍴CR
              O(0=⍴Z)↓cr2,'*** ',(⍕m),' common objects'
              head←Z ⋄ Z←''
     ⍝
     ⍝ we check each object even if SHOW[3]<2
              :For i :In ⍳m
                  Path←fullpath,n←obj[i;]~' '
                  c1←p1.⎕NC n ⋄ c2←p2.⎕NC n
                  :If diff,←c2≠c1 ⍝ same class?
                      O show[3]/cr2,'∘ ',Path,' has different classes'
                  :ElseIf c1∊2 9
                      s1 s2←p1 p2{6::'is undefined' ⋄ ⍺⍎⍵}¨⊂n
                      :If c1=9
                      :AndIf diff[i]←p1 p2≢s1.## s2.## ⍝ children?
                          O show[3]/cr2,'∘ space ',Path,CR,'* External space(s)'
                      :ElseIf diff[i]←0<⍴n1←1↓s1 compvar s2
                          O show[3]/cr2,'∘ variable ',Path,CR,v∆m n1
                      :EndIf
                  :ElseIf c1∊3 4
                      :If diff[i]←0<⍴n1←(p1.⎕CR n)compfn p2.⎕CR n
                          O show[3]/cr2,'∘ function ',Path,CR,n1
                      :EndIf
                  :Else
                      O'∘ ',Path,' has an unknown class: ',⍕c1
                  :EndIf
              :EndFor
              :If show[6]
                  head,←CR,v∆m fit(diff\'≠'),obj
              :EndIf
              any∨←∨/diff
              d←(any\'≠'),head,Z
              Path←pathcopy ⍝ restore for caller
          :EndIf
      :EndIf
    ∇

    FS←⎕SE.SALTUtils.FS

    ∇ r←compare args;tmpa;tmpb;w1;w2;b
     ⍝ danb 2008 ws comparison pgm made into a UCMD in 2010
     ⍝ To compare 2 wss we copy each into a namespace
     ⍝ NOTE: the wss to compare shouldn't have a stack pending for ⎕CY to work.
     
      'At least ONE wss must be specified'⎕SIGNAL 11↓⍨⍴args.Arguments
      ⎕IO←1
      (w1 w2)←¯2↑0,rlb¨args.Arguments
      :If 0≡w1 ⋄ w1←⊂⎕WSID ⋄ :EndIf ⍝ only 1 arg means "compare to this ws"
     
      :If '='=1↑w2 ⍝ same directory? (we allow 1st '=' to mean "same place")
          w2←((-⊥⍨~w1∊':',FS)↓w1),2↓w2 ⍝ remove folder sep too
      :EndIf
      :If ∨/b←('=',FS)=¯1↑w2 ⍝ same filename? (we allow last '=' to mean "same name")
          w2←((-b[1])↓w2),(-⊥⍨~w1∊':',FS)↑w1
      :EndIf
     
      :Trap 11
         ⍝ 1st ws - Create a temp namespace to store the ws to compare (get SYS VARS FIRST!)
          tmpa←⎕NS'' ⋄ SYSVARS tmpa.⎕CY w1 ⋄ tmpa.⎕CY w1 ⋄ tmpa._SystemVars←tmpa⍎,' ',SYSVARS
         ⍝ 2nd ws
          tmpb←⎕NS'' ⋄ SYSVARS tmpb.⎕CY w2 ⋄ tmpb.⎕CY w2 ⋄ tmpb._SystemVars←tmpb⍎,' ',SYSVARS
      :Else
          ')COPY error (pending stack?)'⎕SIGNAL 101
      :EndTrap
      args.(w1 w2)←(w1 w2)
     ⍝ Do it:
      r←args compnss tmpa tmpb
    ∇

    numOf←{v/⍨(,1)≡1⊃(v v)←⎕vfi ⍵}

    ∇ R←compfiles arg;f1;f2;keys;fils;lim;t1;t2;s1;s2;c;t;k1;k2;⍙lim;∆PATH;b;add;key;i;to
     ⍝ compare 2 files as file1 file2. keys optional.
     ⍝ if both args are character they are both filenames
     ⍝ optionally followed by :passnumber
     ⍝ Ex: ]cfcompare 'abc:1234' def -cpts=3 8'
     ⍝     ]cfcomp   33  789 ⍝ tie numbers accepted
      k1←k2←0
      key←{(⍵,'xx')[1+i←i⍳1⌈⌈/i←+\':'=⍵]∊⎕D:(¯1↓i↑⍵)(fi i↓⍵) ⋄ ⍵ 0}
      ((f1 k1)(f2 k2))←key¨arg.Arguments ⍝ arg is 2 files
      :If 0∊⍴t1←numOf f1 ⋄ t1←k1 open f1
      :Else ⋄ f1←trim,⎕FNAMES[⎕FNUMS⍳t1;]
      :EndIf
     ⍝ Accept = as part of the filename to mean "same path or filename"
      :If ∨/b←'='=f2
         ⍝ Find portion to replace '=' by:
          add←1↓⊃(⍲\2⍴1⍴b)/(1,⌽<\⌽f1=FS)⊂';',f1
          :If 1↑b ⋄ f2←add,1↓f2
          :Else ⋄ f2←(¯1↓f2),add
          :EndIf
      :EndIf
      :If 0∊⍴t2←numOf f2 ⋄ t2←k2 open f2
      :Else ⋄ f2←trim,⎕FNAMES[⎕FNUMS⍳t2;]
      :EndIf
      R←'Comparing file <',f1,'>',CR,'          with <',f2,'>',CR
      s1←fsize t1,k1 ⋄ s2←fsize t2,k2
      R,←f1,' has ',(⍕--/t),' components starting at ',⍕1↑lim←t←2↑s1
      R,←CR,f2,' has ',(⍕--/t),' components starting at ',⍕1↑t←2↑s2
      to←{⍺←1 ⋄ ⍺,c1+⍳(1↑⍵)-c1←¯1↑⍺}
      lim[1]⌈←1↑t ⋄ lim[2]←¯1+lim[2]⌊1↓t ⋄ lim←↑to/lim
     ⍝ Are there restrictions on the components to compare?
      :If 0≢t←arg.cpts
          'Invalid component to compare'⎕SIGNAL 11↓⍨∧/t∊⎕D,' ,-'
          t[b/⍳⍴b←t='-']←⊂' to ' ⋄ t←(∊t),',⍬'
          :Trap 0
              lim←lim∩⍎t
          :Else
              'Invalid component to compare'⎕SIGNAL 11
          :EndTrap
      :EndIf
      R,←CR,'Comparing ',(⍕⍴lim),' components'
      c←(2⍴CR),'⍟⍟⍟ Component '
      :For i :In lim
          :If 0<⍴t←1↓(⎕FREAD t1,i,k1)compvar(⎕FREAD t2,i,k2)
              R,←c,⍕i,CR,,t,CR
          :EndIf
      :EndFor
      R,←CR,'⍟⍟⍟ Comparing access matrices'
      R,←t←,CR,(10 0⍕getamof t1,k1)compcm(10 0⍕getamof t2,k2)
      R,←(0∊⍴t)⌿' (no difference)'
      ⎕FUNTIE{⍵/⍨0≡¨≡¨⍵}t1 t2
    ∇

     ⍝ b←→boolean matrix b   m←→ is b with some of the 1's replaced by 0
     ⍝ m has the properties that:(1)   1≥(+/m),+⌿m
     ⍝ (2)  each 1 is below and to the right of the previous 1
     ⍝ (3) +/,m is a maximum for constraints (1) and (2)
     ⍝ NOTES ⎕io←0,   m is not necessarily the unique optimal solution of (3)
     ⍝ main algorithm calculates the boolean matrix x (line l1) where :
     ⍝ max[i;j]←+/,compbool (i,j)↑b    and   x←max-0 1↓0,max
     ⍝ then uses this to calculate result at line l4
    ∇ m←compbool b;br;bc;r;c;x;y;z;i;li;s;fix;n;⎕IO;f
      b←(br←∨/b)⌿(bc←∨⌿b)/b ⍝ get rid of zero rows and columns
      m←(⍴b)⍴r←c←⎕IO←0
     diag:n←⍳+/∧\,0 0⍉b ⍝ follow diagonal as far as possible
      m[r+n;c+n]←n∘.=n ⋄ r←r+n←⍴n ⋄ c←c+n ⋄ b←(n,n)↓b
      x←+/∧\∧/~b ⋄ y←+/∧\∧⌿~b ⍝ number of adjacent 0 rows and cols
      →(0=⌈/x,y)⍴calc ⍝ no zero rows or cols
      r←r+x ⋄ c←c+y ⋄ b←(x,y)↓b ⋄ →diag ⍝ cut back b and try diagonal again
     calc:→(0∊⍴b)⍴end
      b←(fix←i,~i←</⍴b)⍉b ⍝ ensure 1st dimension is longer
      li←⍴y←∨\b[;0] ⋄ z←(1↓⍴b)⍴i←0 ⋄ x←(⍴b)⍴0 ⍝ initialise loop
     l1:x[i;]←z≠¯1↓0,z←⌈\z⌈y[i],(¯1↓z)+1↓b[i;] ⋄ →(li>i←i+1)⍴l1
      →(0=y←¯1↑z)⍴end ⍝ y is the number of matches
      z←n←1↓⍴b ⋄ i←¯1+''⍴⍴b ⍝ initialise loop
     l4:f←z≠s←(z↑b[i;]×+\x[i;])⍳y
      x[i;]←(-s)⊖n↑f ⋄ z←(¯1*f)↑z,s ⋄ i←i-1 ⋄ →(0≠y←y-f)⍴l4
      x[⍳i+1;]←0
      x←b←fix⍉x
      s←⍴b ⋄ m[r+⍳s[0];c++⍳s[1]]←b
     end:m←br⍀bc\m
    ∇

    ∇ d←{m1}compcm m2;x;r;c;⎕IO;del;fmt;w
     ⍝ character matrix comparison
     ⍝ display the object m1 together with the lines from m2 which are
     ⍝ different from m1.
     ⍝ NOTE   2=(⍴⍴m1)=⍴⍴m2
      :If TRIM
          m1←ljust m1 ⋄ m2←ljust m2
      :EndIf
      ⎕IO←0 ⋄ x←compbool m1 jdmr1 m2 ⋄ r←∨/x ⋄ c←~∨⌿x
      del←(~r)\''⍴DELINS ⍝ delete/insert symbols
      fmt←'⍞[⍞,Q⍞]⍞LI',⍕4+99<1↑⍴m1
      w←1↓⍴d←del,fmt ⎕FMT⍳x←⍴r ⍝ row nos, 6 wide, ←=left, →=right
      r←⍋(compind r),c/compind~c ⍝ order of display
      x←m1 cat1 c⌿m2 ⋄ d←d⍪((+/c),w)⍴w↑1↓DELINS ⋄ x←d←(d,x)[r;]
      x←(⍳1+2×ZONE)-ZONE ⍝ zone of display
      r←r/⍳⍴r←d[;⍳2]∨.=2↑1↑DELINS ⋄ r←∪(≠/r∘.<0,1⍴⍴d)/r←,r∘.+x
      d←d[r;]
    ∇

    ∇ R←arg compnss(s1 s2);z;⎕IO;n1;n2;r1;r2;i;j;c;n;m;lim;show;p1;p2;obj;t;x;skip;dis;dif;⎕TRAP;f;Path;tie;⍙result;O;t1;t2;⍙nssrc;⎕ML;nam;c1;c2;d2;d1
     ⍝ DanB 840701 ws comparison using namespaces. arg is switches, sX are nss to compare
     ⍝
     ⍝ SHOW: 1=show names ;2=contents ;3=both ;0=nothing
      show←,2 2⊤3⍴SHOW ⋄ ⎕ML←⎕IO←1
     ⍝ The gathering fn
      R←0⍴⎕FX'O x'((arg.gatheroutput/'R←R,,CR,⍕'),'x')
     
     ⍝ 1- find all obj names in each ws;
     ⍝ nX is the names of a ws, tX their types, rX the number of names
      r1←⍴t1←'VFFP?'[2 3 4 9⍳s1.⎕NC n1←s1.⎕NL 2 3 4 9]
      r2←⍴t2←'VFFP?'[2 3 4 9⍳s2.⎕NC n2←s2.⎕NL 2 3 4 9]
     
     ⍝ names have been found
      O'* comparing ',arg.w1
      O'       with ',arg.w2
      O CR,(t>0)/'NOTE: total sizes differ by ',(⍕t←|-/⎕SIZE 2 2⍴'s1s2'),' bytes.'
     ⍝
     ⍝ 2- display objects not appearing in the other ws:
      obj←' outstanding objects in '
     ⍝
     ⍝ Show objects in the 1st ns NOT in the 2nd
      skip←0       ⍝ Some objects may have been screened out
      :If show[1]>0≡x←arg.exnames ⋄ t←'~'=1↑x ⋄ skip←t≠(t↓x)manymatch n1 ⋄ :EndIf
      :If ∨/i←~c1←×n2 rowfind n1 ⋄ skip←skip∨~show[1]
        ⍝ We display the names even tho we may want to skip their display
          O CR,'*',(4 0⍕+/i),obj,arg.w1     ⍝ show the # of objects
          :If show[4] ⋄ O fit i⌿n1 ⋄ :EndIf ⍝ show their names?
          :For j :In (skip<i)/⍳⍴i ⋄ O s1 displayfp x←(t1,n1)[j;] ⋄ :EndFor ⍝ Some objects have to be shown
      :EndIf
     ⍝ Then the objects in the 2nd ns not in the 1st
      skip←0
      :If show[2]>0≡x←arg.exnames ⋄ t←'~'=1↑x ⋄ skip←t≠(t↓x)manymatch n2 ⋄ :EndIf
      :If ∨/i←~c2←×n1 rowfind n2 ⋄ skip←skip∨~show[2]
          O CR,'*',(4 0⍕+/i),obj,arg.w2
          :If show[5] ⋄ O fit i⌿n2 ⋄ :EndIf
          :For j :In (skip<i)/⍳⍴i ⋄ O s2 displayfp x←(t2,n2)[j;] ⋄ :EndFor
      :EndIf
     
     ⍝ 3- show objects common to both ws
      z←'' ⋄ skip←' '≠dif←' '⍴⍨lim←1⍴⍴obj←(c1/t1),c2⌿t2,n2
      :If 0<lim
          O CR,'*',(4 0⍕lim),' common objects',CR
     ⍝ if ANY output (name or diffs) is required we scan the whole thing
      :AndIf ∨/show[6 3]
          :If 0≢x←arg.exnames ⋄ t←'~'=1↑x ⋄ skip←t≠(t↓x)manymatch 0 2↓obj ⋄ :EndIf
     
     ⍝ 4- display difference between both version of each obj
          :For i :In ⍳lim
              dis←show[3]∧~skip[i]
              m←1⌽'""',nam←2↓obj[i;]~' '
              :If ≠/t←2⍴n←obj[i;] ⋄ dif[i]←'≠' ⍝ same type?
                  z←z,dis/CR,CR,'> ',m,' has different classes (',t[1],' and ',t[2],')'
              :Else
                  ⎕EX 2 2⍴'p1p2' ⍝ external classes cause a problem
                  :If 'FO'∨.=1↑t ⋄ (p1 d1)(p2 d2)←(s1 s2).{(⎕CR ⍵)(0 1 0 1/⎕AT ⍵)}⊂nam
                      r1←compfnnts p1 p2,fmtTS¨d1 d2 ⋄ x←'> Function '
                  :Else
                      p1←s1⍎Path←nam ⋄ p2←s2⍎nam ⋄ x←'> Variable '
                      :If ∨/9≠⎕NC 2 2⍴'p1p2'
                      :OrIf s1 s2≡p1.## p2.##
                          r1←p1 compvar p2
                      :Else
                          r1←'≠External spaces'
                      :EndIf
                  :EndIf
                  :If '≠'=1↑r1 ⋄ dif[i]←'≠' ⋄ z←z,(dis>r1≡'≠')/CR,CR,x,nam,CR,1↓r1 ⋄ :EndIf
              :EndIf
          :EndFor
      :EndIf
     
      :If show[6] ⋄ O fit dif,0 2↓obj ⋄ :EndIf
      O z
    ∇

    ∇ diff←sns comp2fn arg
      diff←⊃compfn/sns.⎕CR¨arg.Arguments
    ∇

    ∇ c←{f1}compfn f2;b;s;n
     ⍝ danb 84/5/30 - compare 2 fns; 199804 don't include String
     ⍝ 2003 use global 'EXSTRING' to screen out the result
     ⍝
      :If ∧/{(isChar ⍵)∧1∊⍴⍴⍵}¨f1 f2 ⋄ (f1 f2)←⎕CR¨f1 f2 ⋄ :EndIf
      (f1 f2)←⎕FMT¨f1 f2
      :If 0≢s←EXSTRING ⋄ n←'~'∊1↑s ⋄ s←n↓s ⍝∇⋄
          s←n≠(1∊s⍷f1)∨1∊s⍷f2
      :EndIf
      c←0,1↓(⍴f1)⌈⍴f2 ⋄ f1←(c⌈⍴f1)↑f1 ⋄ f2←(c⌈⍴f2)↑f2
     ⍝
     ⍝ drop last comment if NOLASTLINE set
      →NOLASTLINE↓l10
      f1←(2↑-'⍝'∊¯1 1↑f1)↓f1 ⋄ f2←(2↑-'⍝'∊¯1 1↑f2)↓f2
     
     ⍝ left justify if so desired
     l10:→TRIM↓l15
      f1←ljust f1 ⋄ f2←ljust f2
     
     ⍝ drop SALT pathname tags
     l15:→(0∊⍴f1)⍴l20
      b←~∨\'⍝∇⍣'⍷,c←f1[1↑⍴f1;] ⋄ f1←(¯1 0↓f1)⍪(1↑b)⌿b\b/c
     l20:→(0∊⍴f2)⍴l30
      b←~∨\'⍝∇⍣'⍷,c←f2[1↑⍴f2;] ⋄ f2←(¯1 0↓f2)⍪(1↑b)⌿b\b/c
     l30:→(f1≡f2)⍴⍴c←''
      c←'≠' ⋄ →s⍴0
      c←v∆m f1 compcm f2
    ∇

    ∇ c←compfnnts(f1 f2 ts1 ts2);b;s;n;rmtag
     ⍝ Compare 2 fns with TimeStamp; 199804 don't include String
     ⍝ c←'' if f1≡f2, '≠...' if they differ,
     ⍝ c←'≠' if they differ but are excluded from being displayed
     
     ⍝ Drop SALT pathname tags
      rmtag←{⍵⌿⍨~('⍝∇⍣'⍷(+/∧\⍵=' ')⌽⍵,' ')[;⎕IO]}
      (f1 f2)←rmtag∘⎕FMT¨f1 f2
     
     ⍝ We do the timestamp/user first to allow skipping over users with EXSTRING
      :If ~EXTS
          f1 f2←⎕FMT¨f1 f2 cat1¨'⍝',¨ts1 ts2
      :EndIf
      :If 0≢s←EXSTRING ⋄ n←'~'∊1↑s ⋄ s←n↓s ⍝∇⋄
          s←n≠(1∊s⍷f1)∨1∊s⍷f2 ⍝ s: SKIP this object
      :EndIf
      c←0,1↓(⍴f1)⌈⍴f2 ⋄ f1←(c⌈⍴f1)↑f1 ⋄ f2←(c⌈⍴f2)↑f2
     
     ⍝ Left justify if so desired
      :If TRIM
          f1←ljust f1 ⋄ f2←ljust f2
      :EndIf
      →(f1≡f2)⍴⍴c←''
      c←'≠' ⋄ →s⍴0
      c←c,v∆m f1 compcm f2
    ∇

    dlb←{(∨\' '≠⍵)/⍵}

    ∇ b←a jdm w;i;n
     ⍝ jot-dot-match with possible trim
      :If TRIM ⋄ a w←⌽∘dlb⍣2¨¨a w ⋄ :EndIf
      b←a∘.≡w
    ∇

    ∇ d←m1 compcv m2;x;r;c;del;fmt;n;b
     ⍝ Enclosed Character Vectors comparison
     ⍝ Display the object m1 together with the lines from m2 which are
     ⍝ different from m1.
      x←compbool m1 jdm m2 ⋄ b←~r←∨/x ⋄ c←~∨⌿x
      del←b\''⍴DELINS ⍝ delete/insert symbols
      fmt←'⍞[⍞,Q⍞]⍞LI',⍕6⌈3+⌊10⍟1⌈(⍴r)-1⍳⍨⌽r
      n←¯1↑⍴d←del,fmt ⎕FMT⍳x←⍴r ⍝ row nos, N(≥6) wide, ←=left, →=right
      r←⍋(compind r),c/compind~c ⍝ order of display
      b←b,1⍴⍨+/c
      d←((↓d),¨m1),(n↑1↓DELINS)∘,¨c/m2 ⋄ d←d[r]
      x←(⍳1+2×ZONE)-ZONE ⍝ zone of display
      r←r/⍳⍴r←b[r] ⋄ r←∪(≠/r∘.<0,1⍴⍴d)/r←,r∘.+x
      d←∊d[r],⍨¨⎕UCS 1↑13 133∩⎕AVU
    ∇

    ∇ n←compind b
     ⍝ return nos. from 1 to +/b for each element of b such that
     ⍝ n is strictly increasing and b/n is ⍳+/b(⎕io=1)
      n←¯1↓+\n/÷n←(1↓n)-¯1↓n←b/⍳⍴b←1,b,1
    ∇

    ∇ diff←src comp2var arg
      diff←{('≠'>.=2↑⍵,'≠')↓⍵}⊃compvar/src⍎¨arg.Arguments
    ∇

    ∇ c←{v1}compvar v2;⎕IO;i;n;ot;o12;s;s1;s2;t;maxw;d;encl;vtvs;AEI;⎕ML;FR
     ⍝ danb 84/5/30 - compare 2 variables. c←'' if v1≡v2, '≠...' if they differ,
     ⍝ c←'≠' if they differ but are excluded from being displayed
     ⍝
      ⎕ML←1 ⋄ FR←⎕IO←0 ⋄ c←0⍴o12←2 3⍴'1st2nd'
      :If ≠/ot←(type v1),type v2
          c←'≠objects are ',⊃{⍺,' and ',⍵}/typeOf¨ot
      :ElseIf ot[1]∊'ps'
     ⍝ same type: package or not
          c←('*** ',(typeOf 1↑ot),' ***')addDiff c←v1 compack v2
      :Else
          :If FR←'r'∊ot
              v1←⎕FMT v1 ⋄ v2←⎕FMT v2
          :EndIf
          →exit if v1≡v2 ⍝ do we have a match?
     ⍝ From here on we take into account the exclusion string:
          n←'~'=1↑t←EXSTRING ⋄ c←'≠' ⋄ →chkr if 0≡t
          t←n↓t
          →(n≠(1∊t⍷⍕v1)∨1∊t⍷⍕v2)⍴exit ⍝ exit if the constraint is filled
     chkr:→((n←⍴⍴v1)=t←⍴⍴v2)⍴sr
          c←'objects differ in rank: ',(⍕n),' and ',⍕t
          →exit
     sr:  n←''⍴n ⋄ t←'' ⍝ same rank
          :If ~i←(⍴v1)≡⍴v2
              c←'objects differ ',t←'in shape (',(⍕⍴v1),' and ',(⍕⍴v2),')'
          :EndIf
     ⍝ OK, let's try to provide a bit more info on the differences.
     ⍝ Let's cover the VTV case right now
          vtvs←(isVTV v1)∧isVTV v2 ⋄ encl←'e'=ot
          →enc if vtvs<encl[0]∧i ⍝ check each element if enclosed and same shape
          →sc if=/ot
          c←'objects differ ',t,(6×0=⍴t)↓'  and in type'
          →exit
     ⍝
     sc:  c←'these are different objects of rank ',⍕n
          →(2<n)⍴exit
          c←'objects are ⎕OR of fns' ⋄ →FR/mat
          t←~i←1↑encl ⋄ i←i/' of enclosures' ⍝∇⋄
          t←t/(5-9×'c'∊ot)↓'char num '       ⍝∇⋄
          c←'objects are ',t,(3 8⍴'scalars vectors matrices')[n;],i
          →(scl,vec,mat)[n]
     scl: c←c,CR,⍕v1,v2 ⋄ →exit
     mat: c←(0≠⍴n)⌿c,CR,n←v∆m(⎕FMT v1)compcm ⎕FMT v2 ⋄ →exit
     vec: n←0
          :If t←(CR∊v1)∧CR∊v2 ⍝ can we make these into matrices?
     ⍝ Would these be fns by any chance (or logos scripts if start w [1])?
              n←((3↑v1)≡n)∧(3↑v2)≡n←'[1]' ⍝ script?
              t←¯1*n←n∨∧/∨/'∇'=2 6⍴(6⍴v1),6⍴v2 ⍝ t=1:object are text lists, ¯1:fn reps
              s1←⍴v1←cut CR,v1 ⋄ s2←⍴v2←cut CR,v2
          :EndIf
          :If (isVTV v1)∧isVTV v2
              c←'objects are ',(¯1 0 1⍳t)⊃'fn representations' 'VTVs' 'text'
              t←5×t=¯1
              c←(0≠⍴n)⌿c,CR,n←(t↓v1)compcv t↓v2 ⋄ →exit
          :EndIf
          →((⍴v1)≡⍴v2)↓end
          c←c,CR,'elements ',(⍕1+i/⍳⍴i←v1≠v2),' are differents (⎕io=1)' ⋄ →end
     ⍝
     enc: AEI←c←'*** arrays of ',v1{0∊⍴⍴⍺:1↓⍵ ⋄ ⍵,⍨' '~⍨⍕{⍺'×'⍵}/⍴⍺}' enclosed items' ⍝ each ∊ is checked
          →(1=n)⍴evc
         ⍝ Let's see if these are ⎕ORs of fns or ns:
          :If (n=0)∧2≡≡v1
          :AndIf ∧/isor¨v1,v2
              t←'of fns' '' '' 'of NSs'⊃⍨d←2⊥'#'=⍬∘⍴¨⎕FMT¨v1,v2
              c←c,' of ⎕ORs ',t ⋄ →d↓mat
             ⍝ Namespaces. See what's in it
              c←1↓c addDiff(⊃v1)compack⊃v2 ⋄ →exit
          :EndIf
          c←c,' of rank ',(⍕n),(1<⍴,v1)/'; they differ at ',(⍕+/~,⊃v1≡¨v2),' places'
          →evc if 0<n
         ⍝ Scalar case
          →end if 0≡≡v1 ⍝ go show the diffs if simple
          :Trap 11 ⍝ this will fail if one is a ⎕OR
              c←1↓c addDiff(⊃v1)compvar⊃v2
          :EndTrap
          →exit
     evc: i←0 ⋄ n←⍴v1 ⋄ c←''
          :For i :In ⍳n
              t←s←''
              :Trap 11 ⍝ this will fail if one is a ⎕OR
                  s←(⊃v1[⊂i])compvar⊃v2[⊂i]
              :Else
                  s←(v1[⊂i])compvar v2[⊂i]
              :EndTrap
              d←+/⎕SI∊⊂'compvar'
              t←' ',d⍴'-' ⍝ show some nesting...
              c←c,(''≢s)/CR,CR,t,' element ',(⍕i+1),CR,v∆m 1↓s
          :EndFor
          c←c,⍨(0<⍴c)/AEI
          →exit
     end: ⍝ Display variables - try to be smart: use ⎕pw to fit the output
          n←¯1↑⍴v1←displayencl v1 ⋄ n←n,¯1↑⍴v2←displayencl v2
          maxw←⎕PW⌊¯1↑MAXDIMS
          →exit if maxw<s←⌈/n ⍝ doesn't fit, exit
     
     ⍝ We can now display columnwise or 2×2 or 1×4 (with IDs)
          →e2x2 if maxw<11++/n
     ⍝ 1×4 works
          n←'Var1 'cat2 v1 cat2' Var2 'cat2 v2 ⋄ →chkd
     e2x2:→e4x1 if maxw<1++/n⌈4
          n←('Var1'cat1 v1)cat2' ','Var2 'cat1 v2 ⋄ →chkd
     e4x1:n←'Var1 'cat1 v1 cat1'Var2 'cat1 v2 ⋄ →chkd
     chkd:
          →exit if(1↑MAXDIMS)<1↑⍴n ⍝ display of variables ?
          c←c,CR,v∆m n
     exit:c,⍨←'≠'⍴⍨×⍴c
      :EndIf
    ∇

    ∇ ∆00←{∆01}default ∆02;∆Ex;∆Rs
     ⍝ DanB 2001 default var ∆01 by value ∆02. '←' in name returns
     ⍝ the value instead and '⍎' will use ⍎ of ∆02 instead of ∆02 itself.
     ⍝ Right arg cannot be a package
      ∆Rs←'←'∊∆01 ⋄ ∆Ex←'⍎'∊∆01 ⋄ ∆01←∆01~'⍎←'
      ∆Ex←2⌊∆Ex+⎕NC ∆01
      ⍎(,⊃1↑∆Rs↓∆01'∆00'),'←',,⊃1↑∆Ex↓'∆02'∆02 ∆01 ⍝
    ∇

    ∇ c←{val}display obj;sep;l;d;s;o;⎕TRAP;ps_;cl;n;ty
     ⍝ danb 840629 display object with value
      ⎕TRAP←(1 10)'C' '→wf' ⍝ cover LIMIT error
     ⍝ 199805 modified to trap ws full
      c←0 0⍴sep←2⍴CR
      cl←1↑obj ⋄ obj←1↓obj~' '
      →(var,var,var,fn,fn,fi,donno)['VPNFOI'⍳cl]
     ⍝
     var:→('p'=ty←type val)⍴pak ⋄ c←((l≢⍬)/' ⍴',⍕l←⍴val),'  ≡',⍕{16::'⍬' ⋄ ⍕≡⍵}val
      c←c,CR,v∆m displayencl val
      ty←1⌽'] [',,(7 5⍴'num  char encl frep null classnamsp')['ncerlsp'⍳ty;]~' '
      c←sep,obj,ty,c ⋄ →0
     wf:c←sep,obj,' [*ws full*]' ⋄ →0
     pak:c←sep,obj,' [',PTYPE,']' ⋄ →0
     fn:c←sep,o,(0∊⍴o←fmtfn⍕val)/'⍫ ',obj ⋄ →0
     fi:c←sep,obj fnfd val ⋄ →0
     donno:c←sep,obj,' is unknown!'
    ∇

    ∇ r←{max}displayencl obj;sr;b;⎕IO
     ⍝ display obj with lines surrounding boxes
      :Trap ⎕IO←1
          :If isor obj
              r←⎕FMT obj
          :Else
              r←disp obj
          :EndIf
          'max'default NLINES,⎕PW
          :If ∨/b←max<sr←⍴r
              r←(sr⌊max)↑(sr⌊max-1 3)↑r ⋄ r[b[1]↑⍴r;⍳3]←r[⍳b[2];¯3↑⍳¯1↑⍴r]←'.'
          :EndIf
          :If 0∊⍴r ⋄ r←⎕FMT'(empty)' ⋄ :EndIf
      :Else
          r←⎕FMT'*** Object is too complex/big'
      :EndTrap
    ∇

    ∇ c←{cl}displayfp obj;typ
     ⍝ danb 840629 display object in CLuster (ns)
      typ←'VFFV?'[2 3 4 9⍳cl.⎕NC obj]
      →('V?'∊typ)/l10,l90
      cl←⎕FMT cl.⎕CR obj ⋄ →l90
     l10:cl←cl⍎obj
     l90:c←cl display typ,obj
    ∇

    ∇ r←{drop}dropsection r;b;⎕IO
     ⍝ drop dot sections from enclosed r
      →drop↓⎕IO←0
      r←r↓⍨¨1+(+\¨'.'=r)⍳¨+/drop ⍝ ensure scalar
    ∇

    ∇ r←fi v
      r←(1+⎕IO)⊃⎕VFI v ⍝
    ∇

    ∇ mat←fit mat;⎕IO;w;p;f;dim
     ⍝ danb 840614 reshape char mat to fit ⎕pw
     ⍝ w:width; f:no of folds; p:page length
      :If ''≡mat
          mat←0 0⍴''
      :Else
          mat←(0,-f⊥f←' '∧.=mat)↓mat ⍝ drop last blank columns
          f←⌊⎕PW÷w←1↓dim←0 2+⍴mat
          dim[⎕IO←0]←f×p←⌈(1⍴dim)÷f
          mat←(p,f×w)⍴1 0 2⍉(f,p,w)⍴dim↑mat
      :End
    ∇

    ∇ r←fmtfn cr;⎕IO;f
     ⍝ danb 2003 format ⎕cr of fn
      f←'R<  ∇ >P<[>Q<]>BLI5'
      ⎕IO←0 ⋄ r←f ⎕FMT⍳1↑⍴cr←⎕FMT cr
      r←v∆m r,cr
    ∇

    fmtTS←{⍕(⊂'G<9999/99/99 Z9:99:99>'⎕FMT 100⊥6↑⎕IO⊃⍵),1↓⍵}

    ∇ cr←{fnn}fncr obj;n
     ⍝ danb 2003 functions description from a namespace
      :Trap 16
          'n'⎕NS obj
      :Else
          n←obj
      :EndTrap
      cr←(¯2↑1,⍴cr)⍴cr←⍕n.⎕CR fnn
    ∇

    ∇ cr←{fnn}fnfd pak;n
     ⍝ danb 2003 functions description from a namespace
      'n'⎕NS pak
      cr←n.⎕VR fnn
    ∇

    ∇ r←fsize t
      r←⎕FSIZE t
    ∇

    ∇ r←getamof tie;⎕TRAP;er_
     ⍝ DanB 19981008 get Access Matrix of file tied to 'tie'
      (⎕TRAP)←19 'C' '→no'
      r←0 3⍴0
      r←⎕FRDAC tie ⋄ →0
     no:⎕←'*** unable to read access matrix of ',,⎕FNAMES[⎕FNUMS⍳1↑tie;]
    ∇

    hasSource←{16 11::0 ⋄ 1∊⍴⍴⎕SRC ⍵}

    if←/⍨

    ∇ r←a iormatch b;simp ⍝ Ignore ⎕ORs, NS & forms
      :If ¯1=r←a noncematch b
          :If (0∊⍴⍴a)∧simp←1≥|≡a   ⍝ this test is flawed?
              r←=/isForNS¨a b ⍝ both the same type? then ≡
          :ElseIf ~simp
              r←∧/,a iormatch¨b
          :EndIf
      :EndIf
    ∇

    ∇ t←isencl a
     ⍝ danb : is a enclosed ?
      t←1∊807 326∊⎕DR a
    ∇

    ∇ b←isform ns
     ⍝ ns is either a namespace or a form
      →(b←9∊⎕NC'ns')↓0
      b←'Form'≡ns.⎕WG'type'
    ∇

    isor←{(0∊⍴⍴⍵)∧1≡≡⍵}

    ∇ r←ispack p
     ⍝ Refs & ⎕OR of refs
      →(r←326∊⎕DR p)↓0 ⋄ →(r←isor p)↓0 ⋄ r←'#⎕'∨.=1⍴⍕p
    ∇

    isChar←{0::0 ⋄ (10|⎕DR ⍵)∊0 2}

    ∇ b←isForNS ns
     ⍝ ns is either a namespace or a form
      :If b←9∊⎕NC'ns' ⍝ 0: not class 9, 1: class 9, char: its Type
          :Trap 6
              b←'FN?'['Form' 'Namespace'⍳⊂nsType ns] ⍝ issue 2764
          :EndTrap
      :Else
      ⍝ It can still be a ⎕OR
          :If isor ns ⋄ b←'O' ⋄ :EndIf
      :EndIf
    ∇

    isVTV←{1≠⍴⍴⍵:0 ⋄ 326≠⎕DR ⍵:0 ⋄ ∧/isChar¨⍵}

    ∇ b←{a}jdmr1 w;i;n
     ⍝ jot-dot-match-rank 1 for rank 2 a and w with same last ⍴
     ⍝ sort of a∘.(≡ö1) w
      i←⍋b←a cat1 w
      i[i]←+\∨/b≠¯1⊖b←b[i;] ⋄ b←i[⍳n]∘.=(n←1⍴⍴a)↓i
    ∇

    lev←{⍺}

    ∇ r←ljust m
     ⍝ left justify matrix
      r←(+/∧\' '=m)⌽m
    ∇

    ∇ r←{pat}manymatch list;⎕IO;n
     ⍝ return list of objects matching several patterns in list
      ⎕IO←1
      pat←1↓(r⍲1⌽r←pat∊' ')/pat←' ',pat ⍝ remove excess blanks
      r←0⌿list ⍝ patterns are delimited by blanks
     l10:r←r⍪pat[⍳¯1+n←pat⍳' ']patmatch list ⋄ →(×⍴pat←n↓pat)⍴l10
      r←((⍳1↑⍴r)=r rowfind r)⌿r ⍝ nub
      r←0≠r rowfind list
    ∇

    ∇ r←a noncematch b;⎕CT
     ⍝ Looks like we cannot use a Dfn here...?
     ⍝ Returns ¯1 if a NONCE error occurs
      r←¯1 ⋄ ⎕CT←1E¯12 ⍝ because of machine differences
      :Trap 16 ⋄ r←a≡b ⋄ :EndTrap
    ∇

      nsType←{'***'≡3⍴⍕F←⍵:'ERROR' ⋄ 2::' OR',⍨(2/0 1='#'∊⍕⍵)/'FnNS'
          isVTV ⍵:{(-1-⍵⍳' ')↑⍵}rlb ⎕IO⊃⍵
          9.6=⎕NC⊂,'F':'Ext'
          0::'???'
          'F'⎕WG'type'}

    ∇ z←{key}open f;t;b;s
     ⍝ Ties a file if not already tied; returns the tie number
     ⍝ The result is a vector if the file was already tied.
     ⍝ We first check that the file is already tied
      :Trap 24 ⋄ z←f ⎕FSTIE 0,key ⍝ scalar result
      :Else ⍝ tying failed, it's already tied, find its tie no
          f←f,'  ' ⍝ in case another file is tied with a longer similar name
         ⍝ account for Windows' C:
          t←(f⍷' ',⍨⍣2+⎕FNAMES)[;1,(b←'\'=FS)/3]
          t←t∨.∧⍲\(1+b)⍴∨⌿t[;1] ⍝ prefer col 1
          →(1=⍴z←t/⎕FNUMS)⍴0 ⍝ there should be one only
         ⍝ Otherwise we give up
          z←0
      :EndTrap
    ∇

    ∇ r←{p}patmatch t;⎕IO;b;d;i;j;n;B;M;S
     ⍝ DANB 950202 PATTERN MATCHING A LA dos BUT BETTER
     ⍝ 't' is the text table shrunk as we go. 'r' is resulting matches
     ⍝ normalize pattern like ?*?* to ??* (** is reduced to *)
      ⎕IO←0 ⋄ n[1]←(⍴p←,p)⌈1↓n←⍴r←t ⋄ t←n↑t ⋄ S←'?' ⋄ M←'*'
      p←1↓n[i],0⍴i←⍒b+((1↓i)-¯1↓i←(j/⍳i),i←⍴b)/⌽100×⍳+/j←2=b←'*?'⍳n←'.',p
      b←1,(1↓b)⍲¯1↓b←M=p←,p ⋄ p←b/p
      B←t∊' ' ⍝ B is Blanks, S Single char NOT Blank, M Many S
     ⍝ chk if * only and exit else grab chars up to * and goto many if 1st *
     l:→(⍴p)↓N ⋄ d←p[0]∊M ⋄ →(⍴p←d↓p)↓0 ⋄ n←p⍳M ⋄ b←S≠n↑p ⋄ i←b/⍳n ⋄ →d⍴many
     ⍝ SIMPLE PATTERN: all chars must be non blank and non ? must match
      t←(0,n)↓(d←(∨/B[;⍳n])<t[;i]∧.=b/n↑p)⌿t ⍝∇⋄
      r←d⌿r ⋄ p←n↓p ⋄ →l,⍴B←(0,n)↓d⌿B
     ⍝ MANY: rotate left until next matching string
     many:d←⍴t ⋄ t←,t ⋄ j←b/⍳⍴b←(1-n)↓t∊1↑p
      b[j]←t[j∘.+i]∧.=p[i] ⋄ t←d⍴t
     ⍝ if it is the last pattern we want to check the last not the 1st
      i←∨/b←∨\j←(0,1-n)↓d⍴b ⋄ →(0≠⍴p←n↓p)⍴l10 ⍝∇⋄
      i←∨/b←∨\⌽<\⌽j
     l10:j←⍴n←n++/~i⌿b ⋄ r←i⌿r ⍝∇⋄
      d←j,1↓d ⋄ B←' '=t←d↑n⌽(i⌿t),d⍴' ' ⋄ →l
     ⍝ End of pattern. Keep whatever doesn't have a reminder.
     N:r←(∧/B)⌿r
    ∇

    rlb←{(+/∧\⍵=' ')↓⍵}

    ∇ index←{left}rowfind right;sl;sr;iota;i;⎕IO;n
     ⍝ returns the locations of the rows of right within left for char mats
     ⍝ left and right need not match in width. Not found=0
      ⎕IO←1
      i←0,1↓(sl←⍴left←(¯2↑1 1,⍴left)⍴left)⌈sr←⍴right←(¯2↑1 1,⍴right)⍴right
      n←1+1↑⍴left←(i⌈sl)↑left ⋄ right←(i⌈sr)↑right
      index←n|(↓left)⍳↓right
    ∇

    ∇ {e}shcpt w
      →(⍴w)↓0 ⍝ anything to show?
      'Path'Oval e
      'type: ','type'Oval e
      w,CR
    ∇

    ∇ s←sqz s;w
     ⍝ danb 83/10/12 SQeeZe out extra spaces
      s←1↓(w⍲1⌽w←s∊' ')/s←' ',s
    ∇

    ∇ r←strlen vec;⎕IO
     ⍝ find length of each section where delimiter is 1↑vec
      ⎕IO←0 ⋄ r←(1↓r)-¯1↓r←r/⍳⍴r←1,1⌽vec=1↑vec
    ∇

    ∇ r←tomat obj
     ⍝ Dyadic's version is simple
      :If 3>|≡obj
      :AndIf 3>⍴⍴obj
      :AndIf 100∧.>⍴obj
          :Trap 10
              r←⎕FMT obj
          :Else
              r←⎕FMT'*** LIMIT error ***'
          :EndTrap
      :Else
          r←⎕FMT'** Object too complex **'
      :EndIf
    ∇

    trim←{⍵/⍨(∨\b)∧⌽∨\⌽b←⍵≠' '}

    ∇ r←type o;⎕TRAP;e
     ⍝ one-letter type for Object
      r←'s' ⋄ →(9∊⎕NC'o')⍴0 ⍝ namespace/ref?
      r←'l' ⋄ ⎕TRAP←11 'C' 'e←1⋄→l10×~⎕null≡o' ⋄ e←0 ⋄ r←⎕DR o ⍝ ⎕nulls
     l10:r←'p' ⋄ →e⍴0 ⍝ V10+
      r←'p' ⋄ →(ispack o)⍴0
      r←'r' ⋄ →(isor o)⍴0 ⍝ fn representation (⎕OR)
      r←'e' ⋄ →(isencl o)⍴0
     ⍝ num is broad
      r←'cn'[0 1⍳0≡''⍴1↑0⍴o]
    ∇

    typeOf←{('specnr'⍳⍵)⊃'Ref'PTYPE,'enclosed' 'character' 'numeric' 'fn ⎕OR',¨⊂' array'}

    ∇ v←v∆m m
     ⍝ danb 840531 - vector from matrix (compare grp)
      v←1↓(,1,⌽~∧\' '=⌽m)/,CR,m
    ∇

    ∇ r←NLnoSALT obj;⎕ML
     ⍝ Names in obj w/o SALT
      r←obj.⎕NL-2 3 4 ⋄ ⎕ML←0
      r←↑r,(obj.⎕NL-9)~⊂'SALT_Data'
     ⍝ r←↑r,{m←⍕⍵ ⋄ 0∊⍴a←m∘,¨'.',¨n←⍵.⎕NL-9:⍬ ⋄ n/⍨a≡¨⍕¨⍵⍎¨n}obj
    ∇

    ∇ r←Windows
      r←'Windows'≡7↑⎕IO⊃'.'⎕WG'aplversion' ⍝ should account for Linux
    ∇

      disp←{⎕IO ⎕ML←0
     
          ⍺←⍬ ⋄ opt ldr ctd←3↑⍺,3/1
     
          ul uc ur←ldr⊃⌽'┌┬┐' '.'
          ml mc mr←ldr⊃⌽'├┼┤' '|+|'
          ll lc lr←ldr⊃⌽'└┴┘' ''''
          vt hz←ldr⊃'|-' '│─'
     
          box←{
              isor ⍵:⎕FMT⊂⍵
              1≥|≡⍵:open ⎕FMT open ⍵
              subs←expand ∇¨matr open ⍵
              odec←(shape ⍵)outer subs
              idec←inner ⍵
              (odec,idec)collect subs
          }
     
          expand←{
              rows cols←sepr⍴¨⍵
              sizes←(⌈/rows)∘.,⌈⌿cols
              ctd=0:sizes↑¨⍵
              v h←sepr⌈0.5×↑(⍴¨⍵)-sizes
              v⊖¨h⌽¨sizes↑¨⍵
          }
     
          outer←{
              on←{⍺:⍺⍺ ⍵ ⋄ ⍵}
              sizes←1 0{⊃↓⍺⍉on ⍵}¨sepr⍴¨⍵
              sides←sizes/¨¨vt hz
              bords←1↓¨ml uc glue¨sides
              ↑,¨/(ul'')⍺ bords(ll ur)
          }
     
          inner←{
              deco←{(type ⍵),shape ⍵}
              sepr deco¨matr open ⍵
          }
     
          collect←{
              lft top tt vv hh←⍺
              boxes←vv right(tt hh)lower ⍵
              lft,top⍪↑⍪⌿,/open boxes
          }
     
          lower←{
              bot←{⍵⍪(-1⊃⍴⍵)↑⍺ split ⍵}
              split←{((¯2+1⊃⍴⍵)/hz)glue ⍺}
              (matr↑,¨/⍺)bot¨matr ⍵
          }
     
          right←{
              types←2⊥¨(⍳⍴⍵)=⊂¯1+⍴⍵
              chars←mc mr lc lr[types]
              rgt←{⍵,(¯2+⊃⍴⍵)1 1/vt,⍺}
              ((matr ⍺),¨chars)rgt¨⍵
          }
     
          type←{
              opt<|≡⍵:hz
              isor ⍵:'∇'
              sst←{
                  0=opt×⍴⍴⍵:hz
                  (⊃⍵∊'¯',⎕D)⊃'#~'
              }∘⍕
              0=≡⍵:sst ⍵
              {(1=⍴⍵)⊃'+'⍵}∪,sst¨open ⍵
          }
     
          shape←{
              0=opt×⍴⍴⍵:vt hz
              cols←(×¯1↑⍴⍵)⊃'⊖→'
              rsig←(××/¯1↓⍴⍵)⊃'⌽↓'
              rows←(¯1+3⌊⍴⍴⍵)⊃vt rsig'⍒'
              rows cols
          }
     
          matr←{↑,↓⍵}
          sepr←{+/¨1⊂↑⍵}
          open←{16::(1⌈⍴⍵)⍴⊂'[ref]' ⋄ (1⌈⍴⍵)⍴⍵}
          isor←{1 ⍬≡(≡⍵)(⍴⍵)}
          glue←{↑⍺{⍺,⍺⍺,⍵}/⍵}
     
          box ⍵
      }

:EndNamespace